(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Infeasibility of Conditional Critical Pairs\<close>

theory Infeasibility
imports
  "../Nonreachability/Nonreachability"
  "../Rewriting/Tcap_Impl" (*FIXME: check why order matters for adhoc overloading*)
  Conditional_Critical_Pairs
  Conditional_Rewriting_Impl
  Transitive_Closure_List_Impl
begin

datatype ('f, 'v) infeasibility_proof =
  Infeasible_Compound_Conditions 'f "('f, 'v) nonreachability_proof"
| Infeasible_Equation "('f, 'v) term" "('f, 'v) term" "('f, 'v) nonreachability_proof"
| Infeasible_Subset "('f, 'v) rules" "('f, 'v) infeasibility_proof"
| Infeasible_Rhss_Equal "('f, 'v) term" "('f, 'v) term" "('f, 'v) term" "('f, 'v) nonjoinability_proof"

abbreviation "err_not_cond e \<equiv>
  shows ''equation '' \<circ> shows_eq e \<circ> shows '' is not in list of conditions'' \<circ> shows_nl"

(*TODO: move*)
definition "trancl_of_list xs =
  concat (map (\<lambda>x. map (\<lambda>y. (x, y)) (trancl_list_impl xs [x])) (map fst xs))"

lemma set_trancl_of_list [simp]:
  "set (trancl_of_list R) = (set R)\<^sup>+"
proof (intro equalityI subrelI)
  fix x y assume "(x, y) \<in> (set R)\<^sup>+"
  then show "(x, y) \<in> set (trancl_of_list R)"
    by (induct) (auto dest: trancl_into_trancl simp: trancl_list_impl trancl_of_list_def, force)
qed (auto simp: trancl_list_impl trancl_of_list_def)

lemma conds_sat_trancl_of_list:
  assumes "conds_sat R n cs \<sigma>"
  shows "conds_sat R n (trancl_of_list cs) \<sigma>"
using assms by (fastforce simp: conds_sat_iff elim: trancl.induct)

fun check_infeasible'
where
  "check_infeasible' R cs (Infeasible_Compound_Conditions f p) =
    check_nonreachable (map fst R) (Fun f (map fst cs)) (Fun f (map snd cs)) p"
| "check_infeasible' R cs (Infeasible_Equation s t p) = do {
    check ((s, t) \<in> set cs) (err_not_cond (s, t));
    check_nonreachable (map fst R) s t p
  }"
| "check_infeasible' R cs (Infeasible_Subset cs' p) = do {
    check_subseteq cs' cs <+? err_not_cond;
    check_infeasible' R cs' p
  }"
| "check_infeasible' R cs (Infeasible_Rhss_Equal s t u p) = do {
    check ((s, u) \<in> set cs) (err_not_cond (s, u));
    check ((t, u) \<in> set cs) (err_not_cond (t, u));
    check_nonjoinable (map fst R) s t p
  }"

definition
  check_infeasible ::
    "(('f::{key, show}, 'v::show) rules \<times> ('f, 'v) infeasibility_proof) list \<Rightarrow>
      ('f, 'v) crules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_infeasible css R cs\<^sub>1 cs\<^sub>2 = check_exm (\<lambda>(cs, p). do {
    let cs' = cs\<^sub>1 @ cs\<^sub>2;
    check (match_rules cs cs' \<noteq> None \<and> match_rules cs' cs \<noteq> None) id;
    check_infeasible' R (trancl_of_list cs) p
  }) css (shows_sep id shows_nl)"

lemma rules_enc_aux:
  fixes xs ys :: "('f, 'v :: infinite) rules"
  assumes "map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs = map fst ys @ map snd ys"
  shows "map (op \<bullet> \<pi>) xs = ys"
proof -
  have "length xs = length ys"
  proof (rule ccontr)
    assume "length xs \<noteq> length ys"
    moreover
    have "length (map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs) = length (map fst ys @ map snd ys)"
      using assms by simp
    ultimately show False by simp
  qed
  then show ?thesis
    using assms by (induct xs ys rule: list_induct2; auto simp: eqvt)
qed

lemma perm_term_of_crule_eq_conv:
  "\<pi> \<bullet> term_of_crule \<rho>\<^sub>1 = term_of_crule \<rho>\<^sub>2 \<longleftrightarrow> \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2"
by (cases \<rho>\<^sub>1 rule: crule_cases; cases \<rho>\<^sub>2 rule: crule_cases; auto simp: eqvt rules_enc_aux)

lemma perm_term_of_rules_eq_conv:
  "\<pi> \<bullet> term_of_rules rs\<^sub>1 = term_of_rules rs\<^sub>2 \<longleftrightarrow> \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
by (auto simp: eqvt rules_enc_aux)

lemma match_crule_imp_variants:
  fixes \<rho>\<^sub>1 \<rho>\<^sub>2 :: "('f, 'v :: infinite) crule"
  assumes "match_crule \<rho>\<^sub>2 \<rho>\<^sub>1 = Some \<sigma>\<^sub>1" and "match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 = Some \<sigma>\<^sub>2" 
  shows "\<exists>\<pi>. \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2"
proof -
  have "term_of_crule \<rho>\<^sub>2 \<cdot> \<sigma>\<^sub>2 = term_of_crule \<rho>\<^sub>1"
    and "term_of_crule \<rho>\<^sub>1 \<cdot> \<sigma>\<^sub>1 = term_of_crule \<rho>\<^sub>2"
    using assms by (auto simp: match_crule_alt_def dest: match_sound)
  then obtain \<pi> where "\<pi> \<bullet> \<rho>\<^sub>2 = \<rho>\<^sub>1"
    using term_variants_iff [of "term_of_crule \<rho>\<^sub>2" "term_of_crule \<rho>\<^sub>1"]
    unfolding perm_term_of_crule_eq_conv by auto
  then have "-\<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2" by auto
  then show ?thesis ..
qed

lemma match_rules_imp_variants:
  fixes rs\<^sub>1 rs\<^sub>2 :: "('f, 'v :: infinite) rules"
  assumes "match_rules rs\<^sub>2 rs\<^sub>1 = Some \<sigma>\<^sub>1" and "match_rules rs\<^sub>1 rs\<^sub>2 = Some \<sigma>\<^sub>2"
  shows "\<exists>\<pi>::'v perm. \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
proof -
  have "term_of_rules rs\<^sub>2 \<cdot> \<sigma>\<^sub>2 = term_of_rules rs\<^sub>1"
    and "term_of_rules rs\<^sub>1 \<cdot> \<sigma>\<^sub>1 = term_of_rules rs\<^sub>2"
    using assms by (auto simp: match_rules_alt_def dest: match_sound)
  then obtain \<pi> :: "'v perm" where "\<pi> \<bullet> rs\<^sub>2 = rs\<^sub>1"
    using term_variants_iff [of "term_of_rules rs\<^sub>2" "term_of_rules rs\<^sub>1"]
    unfolding perm_term_of_rules_eq_conv by auto
  then have "-\<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2" by (auto simp del: rules_pt.permute_list_def)
  then show ?thesis ..
qed

lemma check_infeasible':
  assumes "isOK (check_infeasible' R cs p)"
  shows "\<not> (\<exists>\<sigma> n. conds_sat (set R) n cs \<sigma>)"
using assms
proof (induct p arbitrary: cs)
  case ok: (Infeasible_Compound_Conditions f p)
  let ?s = "Fun f (map fst cs)"
  let ?t = "Fun f (map snd cs)"
  have "\<not> (\<exists>\<sigma>. (?s \<cdot> \<sigma>, ?t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*)"
    using ok by (auto simp: Ru_def dest: check_nonreachable)
  moreover
  { fix \<sigma> and n assume "conds_sat (set R) n cs \<sigma>"
    then have "\<forall>(s, t) \<in> set cs. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*"
      using rtrancl_mono [OF cstep_imp_Ru_step, of "set R"]
      by (auto simp: conds_sat_iff split: prod.splits dest!: csteps_n_imp_csteps)
    then have "(?s \<cdot> \<sigma>, ?t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*"
      by (unfold subst_apply_term.simps, intro args_steps_imp_steps)
         (auto intro: args_steps_imp_steps) }
   ultimately show ?case by auto
next
  case (Infeasible_Equation s t p)
  then have "(s, t) \<in> set cs" and "\<not> (\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*)"
    by (auto simp: Ru_def dest: check_nonreachable)
  then show ?case
    using csteps_n_imp_csteps [of _ _ "set R"]
      and cstep_imp_Ru_step [THEN rtrancl_mono, of "set R"] by (auto simp: conds_sat_iff) blast
next
  case (Infeasible_Subset cs' p)
  then show ?case by (auto simp: conds_sat_iff) blast
next
  case (Infeasible_Rhss_Equal s t u p)
  then have "(s, u) \<in> set cs" and "(t, u) \<in> set cs"
    and "\<not> (\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>\<down>)" by (auto simp: Ru_def dest: check_nonjoinable)
  then show ?case
    using csteps_n_imp_csteps [of _ _ "set R"] and cstep_imp_Ru_step [THEN rtrancl_mono, of "set R"]
    by (blast dest: conds_satD)
qed

lemma check_infeasible:
  fixes R :: "('f::{key, show}, 'v::{show, infinite}) crules"
  assumes "isOK (check_infeasible css R cs\<^sub>1 cs\<^sub>2)"
  shows "\<not> (\<exists>\<sigma> n. conds_sat (set R) n (cs\<^sub>1 @ cs\<^sub>2) \<sigma>)"
proof -
  let ?cs' = "cs\<^sub>1 @ cs\<^sub>2"
  obtain \<pi> :: "'v perm" and cs and p where "(cs, p) \<in> set css"
    and cs: "cs = \<pi> \<bullet> ?cs'" and ok: "isOK (check_infeasible' R (trancl_of_list cs) p)"
    using assms and match_rules_imp_variants [of _ ?cs']
    by (fastforce simp: check_infeasible_def Ru_def)
  then have "\<not> (\<exists>\<sigma> n. conds_sat (set R) n cs \<sigma>)"
    using conds_sat_trancl_of_list [of "set R" _ cs] by  (blast dest: check_infeasible')
  then show ?thesis by (metis cs conds_sat_perm_shift rules_pt.permute_minus_cancel(2))
qed

end
