(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2016)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Critical_Pair_Closure_Impl
imports
  Critical_Pair_Closure
  Critical_Pairs_Impl
  "../Framework/QDP_Framework_Impl"
begin

definition check_strongly_closed where
  "check_strongly_closed R n = do {
     check_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
         \<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0))))
       (shows ''the critical pair '' +@+ shows_term s +@+ shows '' <- . -> '' +@+ shows_term t +@+
        shows '' is not strongly closed within '' +@+ shows n +@+ shows '' steps.'')
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s +@+ shows_nl +@+ shows ''hence the following TRS is not strongly closed '' +@+
       shows_nl +@+ shows_trs R +@+ indent s)"

lemma check_strongly_closed:
  assumes "isOK(check_strongly_closed R n)"
  shows "CR (rstep (set R))"
proof (rule strongly_closed_linear_CR)
  from assms show "linear_trs (set R)" unfolding check_strongly_closed_def by auto
  let ?red = "\<lambda>s n. reachable_terms R s n"
  let ?opt = "\<lambda>s t. List.find (\<lambda>x. List.member (?red s n) x) (?red t (Suc 0))"
  { fix s t
    assume "\<not> Option.is_none (?opt s t)"
    then obtain r where r: "?opt s t = Some r" by (cases "?opt s t") auto
    then have "List.member (?red s n) r" using find_Some_iff by metis
    then have mem:"r \<in> set (?red s n)" using in_set_member by fast
    from r have "\<exists>i<length (?red t (Suc 0)). r = (?red t (Suc 0)) ! i" using find_Some_iff by metis
    then have "r \<in> set (?red t (Suc 0))" using nth_mem by blast
    with mem have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      using reachable_terms reachable_terms_one by metis
  } note sound=this
  { fix b s t
    assume "(b, s, t) \<in> critical_pairs (set R) (set R)"
    with assms have *:"\<not> Option.is_none (?opt s t) \<and> \<not> Option.is_none (?opt t s)"
      unfolding check_strongly_closed_def by fastforce
    with sound have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      and  "\<exists>u. (t, u) \<in> (rstep (set R))\<^sup>* \<and> (s, u) \<in> (rstep (set R))\<^sup>=" by blast+
  }
  then show "strongly_closed (set R)" unfolding strongly_closed_def by blast
qed

end
