theory Safety_Checker
imports 
  "$AFP/Check_Monad/Check_Monad"
  "~~/src/HOL/Library/RBT_Mapping"
  Entailment_Checker
  ITS
begin

type_synonym 'v invariant = "'v formula"
type_synonym art_node = string

datatype 't art_edge = is_cover_node: Cover art_node | Children "('t \<times> art_node) list"

record ('v,'l,'t)art = 
  initial_node :: art_node
  art_nodes :: "art_node set"
  art_edge :: "art_node \<Rightarrow> 't art_edge"
  node_location :: "art_node \<Rightarrow> 'l"
  node_invariant :: "art_node \<Rightarrow> 'v invariant"

definition trans_formula_of :: "('v,'l,'t)lin_prog \<Rightarrow> 't \<Rightarrow> 'v transition_formula" where
  "trans_formula_of P \<tau> = trans_formula (transition_of P \<tau>)"

definition transition_entailment :: "('v,'l,'t)lin_prog \<Rightarrow> 'v formula \<Rightarrow> 't \<Rightarrow> 'v formula \<Rightarrow> bool" where
  "transition_entailment P \<phi> \<tau> \<psi> \<equiv> ((map_formula Pre \<phi>) and (trans_formula_of P \<tau>)) implies (map_formula Post \<psi>)"

context
  fixes err :: "'l :: show"
  and P :: "('v :: {compare_order,show},'l, 't :: show)lin_prog"
  and A :: "('v,'l,'t)art"
begin

definition simulation_cond :: bool where 
  "simulation_cond \<equiv> initial_node A \<in> art_nodes A 
    \<and> node_location A (initial_node A) = initial P
    \<and> (\<forall> a tas l \<tau> \<phi> l'. a \<in> art_nodes A 
     \<longrightarrow> art_edge A a = Children tas 
     \<longrightarrow> l = node_location A a
     \<longrightarrow> \<tau> \<in> transitions P
     \<longrightarrow> transition_of P \<tau> = Transition l \<phi> l'
     \<longrightarrow> (\<exists> b \<in> art_nodes A. (\<tau>, b) \<in> set tas \<and> node_location A b = l'))"

definition error_cond :: "bool" where
  "error_cond \<equiv> \<forall> a \<in> art_nodes A. node_location A a = err 
    \<longrightarrow> node_invariant A a implies form_False"

definition cover_edges_cond :: bool where
  "cover_edges_cond \<equiv> (\<forall> a b. a \<in> art_nodes A 
     \<longrightarrow> art_edge A a = Cover b 
     \<longrightarrow> (b \<in> art_nodes A \<and> node_location A a = node_location A b
       \<and> (\<not> is_cover_node (art_edge A b)) 
       \<and> (node_invariant A a implies node_invariant A b)))"

definition initial_cond :: bool where
  "initial_cond \<equiv> node_invariant A (initial_node A) = form_True"

definition children_edges_cond :: bool where
  "children_edges_cond \<equiv> (\<forall> a tas \<tau> b. a \<in> art_nodes A 
     \<longrightarrow> art_edge A a = Children tas
     \<longrightarrow> (\<tau>,b) \<in> set tas
     \<longrightarrow> transition_entailment P (node_invariant A a) \<tau> (node_invariant A b))"

definition well_formed :: bool where 
  "well_formed \<equiv> children_edges_cond \<and> cover_edges_cond \<and> initial_cond 
    \<and> error_cond \<and> simulation_cond"

fun corresponds_to :: "('v,'l)state \<Rightarrow> art_node \<Rightarrow> bool" where 
  "corresponds_to (State \<alpha> l) a  = (node_location A a = l \<and> 
    eval_lin_formula \<alpha> (node_invariant A a))"

lemma pre_valuation: assumes eval: "eval_lin_formula \<alpha>1 \<phi>"
  shows "eval_lin_formula (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (map_formula Pre \<phi>)"
proof (cases \<phi>)
  case (Conjunction \<phi>s)
  show ?thesis using eval unfolding Conjunction eval_lin_formula.simps formula.simps set_map
  proof (induct \<phi>s)
    case (Cons \<phi> \<phi>s)
    hence eval: "eval_lin_atom \<alpha>1 \<phi>" and 
      IH: "\<forall>a\<in>map_lin_atom Pre ` set \<phi>s. eval_lin_atom (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) a" by auto
    obtain exp where \<phi>: "\<phi> = Exp_Le_0 exp" by (cases \<phi>, auto)
    obtain i cs where exp: "exp = Lin_Exp i cs" by (cases exp, auto)
    have [simp]: "(\<Sum>x\<leftarrow>cs. case map_prod Pre id x of (x, c) \<Rightarrow> c * pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta> x)
      = (\<Sum>(x, c)\<leftarrow>cs. c * \<alpha>1 x)"
      by (induct cs, auto)
    from eval have "eval_lin_atom (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (map_lin_atom Pre \<phi>)"
      unfolding \<phi> exp by (simp add: o_def)
    with IH show ?case by simp
  qed simp
qed

lemma post_valuation: assumes eval: "eval_lin_formula (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (map_formula Post \<phi>)"
  shows "eval_lin_formula \<alpha>2 \<phi>"
proof (cases \<phi>)
  case (Conjunction \<phi>s)
  show ?thesis using eval unfolding Conjunction eval_lin_formula.simps formula.simps set_map
  proof (induct \<phi>s)
    case (Cons \<phi> \<phi>s)
    hence eval: "eval_lin_atom (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (map_lin_atom Post \<phi>)"
      and IH: "\<forall>a\<in>set \<phi>s. eval_lin_atom \<alpha>2 a" by auto
    obtain exp where \<phi>: "\<phi> = Exp_Le_0 exp" by (cases \<phi>, auto)
    obtain i cs where exp: "exp = Lin_Exp i cs" by (cases exp, auto)
    have [simp]: "(\<Sum>x\<leftarrow>cs. case map_prod Post id x of (x, c) \<Rightarrow> c * pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta> x)
      = (\<Sum>(x, c)\<leftarrow>cs. c * \<alpha>2 x)"
      by (induct cs, auto)
    from eval have "eval_lin_atom \<alpha>2 \<phi>"
      unfolding \<phi> exp by (simp add: o_def)
    with IH show ?case by simp
  qed simp
qed

  
lemma well_formed_sound: assumes well_formed  
  shows "safety (Safety_Prog err P)"
  unfolding safety_def safety_prog.simps
proof (rule ccontr)
  assume "\<not> (\<forall>\<alpha>. State \<alpha> err \<notin> reachable_states P)"
  then obtain \<alpha> where reach: "State \<alpha> err \<in> reachable_states P" by auto
  from this[unfolded reachable_states_def] obtain init where
    init: "init \<in> initial_states P" and steps: "(init, State \<alpha> err) \<in> (transition P)^*" by auto
  from assms[unfolded well_formed_def] have ic: "initial_cond" and ec: "error_cond"
    and sc: "simulation_cond" and cec: "children_edges_cond" 
    and cc: "cover_edges_cond" by auto
  from sc[unfolded simulation_cond_def]
  have *: "initial_node A \<in> art_nodes A" "node_location A (initial_node A) = initial P" by auto
  {
    note ic = ic[unfolded initial_cond_def]
    obtain l \<alpha> where id: "init = State \<alpha> l" by (cases init, auto)
    with init[unfolded id initial_states_def] have init: "init = State \<alpha> (initial P)" by auto
    have "corresponds_to init (initial_node A)" unfolding init using ic *
      by simp
  } note initial_correspondence = this
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> transition P" and s1a1: "corresponds_to s1 a1" and mem: "a1 \<in> art_nodes A"
    obtain \<alpha>1 l1 where s1: "s1 = State \<alpha>1 l1" by (cases s1, auto)
    obtain \<alpha>2 l2 where s2: "s2 = State \<alpha>2 l2" by (cases s2, auto)
    from t[unfolded transition_def s1 s2] obtain \<tau> \<phi> \<beta>
    where trans: "\<tau> \<in> transitions P" "transition_of P \<tau> = Transition l1 \<phi> l2"
      and transf: "eval_lin_formula (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) \<phi>"
      by blast
    def \<alpha> \<equiv> "(pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>)"
    have "\<exists> a1 tas. corresponds_to s1 a1 
      \<and> art_edge A a1 = Children tas
      \<and> a1 \<in> art_nodes A"
    proof (cases "art_edge A a1")
      case (Children tas)
      with s1a1 mem show ?thesis by blast
    next
      case (Cover b)
      from cc[unfolded cover_edges_cond_def, rule_format, OF mem Cover] obtain tas where
        b: "b \<in> art_nodes A" and loc: "node_location A a1 = node_location A b"
        and edge: "art_edge A b = Children tas"
        and impl: "node_invariant A a1 implies node_invariant A b" by (cases "art_edge A b", auto)
      show ?thesis
      proof (rule exI[of _ b], rule exI[of _ tas], intro conjI)
        from s1a1[unfolded s1, simplified]
        have loc: "node_location A b = l1" and alpha: "eval_lin_formula \<alpha>1 (node_invariant A a1)" 
          using loc by auto
        from impliesD[OF impl alpha] have "eval_lin_formula \<alpha>1 (node_invariant A b)" .
        with loc show "corresponds_to s1 b" unfolding s1 by auto
      qed (insert b edge, auto)
    qed
    then obtain a1 tas where s1a1: "corresponds_to s1 a1"
      and edge: "art_edge A a1 = Children tas"
      and a1: "a1 \<in> art_nodes A" by blast
    from s1a1[unfolded s1] have l1: "l1 = node_location A a1"
      and \<alpha>1: "eval_lin_formula \<alpha>1 (node_invariant A a1)" by auto
    from sc[unfolded simulation_cond_def, THEN conjunct2, THEN conjunct2, 
      rule_format, OF a1 edge l1 trans] obtain b where
      b: "b \<in> art_nodes A" and mem: "(\<tau>, b) \<in> set tas" and l2: "node_location A b = l2" by auto
    from cec[unfolded children_edges_cond_def, rule_format, OF a1 edge mem]
    have "transition_entailment P (node_invariant A a1) \<tau> (node_invariant A b)" .
    from this[unfolded transition_entailment_def trans_formula_of_def] trans(2)
    have "map_formula Pre (node_invariant A a1) and \<phi> implies map_formula Post (node_invariant A b)" by simp
    from impliesD[OF this form_andI[OF pre_valuation[OF \<alpha>1] transf]]
    have "eval_lin_formula (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (map_formula Post (node_invariant A b))" .
    hence "eval_lin_formula \<alpha>2 (node_invariant A b)" by (rule post_valuation)
    with l2 have "corresponds_to s2 b" unfolding s2 by simp
    with b have "\<exists> a2 \<in> art_nodes A. corresponds_to s2 a2" by blast
  } note one_step_simulation = this
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> (transition P)^*" and s1a1: "corresponds_to s1 a1" and mem: "a1 \<in> art_nodes A"
    from t have "\<exists> a2 \<in> art_nodes A. corresponds_to s2 a2" 
    proof (induct rule: rtrancl_induct)
      case (step s2 s3)
      from step(3) obtain a2 where a2: "a2 \<in> art_nodes A" and s2a2: "corresponds_to s2 a2" by blast
      from one_step_simulation[OF step(2) s2a2 a2] show ?case .
    qed (insert s1a1 mem, auto)
  } note many_step_simulation = this
  from many_step_simulation[OF steps initial_correspondence *(1)]
    obtain a_err where a_err: "a_err \<in> art_nodes A" and err: "corresponds_to (State \<alpha> err) a_err" by auto
  from err[simplified] have err: "node_location A a_err = err" "eval_lin_formula \<alpha> (node_invariant A a_err)" by auto
  from ec[unfolded error_cond_def, rule_format, OF a_err err(1)] have "node_invariant A a_err implies form_False" .
  from impliesD[OF this err(2)] show False unfolding form_False_def by simp
qed

context
  fixes ans :: "art_node list"
  and trans_P_list :: "'l \<Rightarrow> ('t \<times> 'l)list"
  and error_hints :: "art_node \<Rightarrow> entailment_info option"
  and cover_hints :: "art_node \<Rightarrow> entailment_info option"
  and transition_hints :: "art_node \<Rightarrow> entailment_info list option"
begin

definition simulation_cond_checker :: "shows check" where
  "simulation_cond_checker \<equiv> do {
      check (initial_node A \<in> art_nodes A) (shows ''initial node of A ('' o shows (initial_node A) 
        o shows '') is not mentioned as node of ART'');
      check (node_location A (initial_node A) = initial P) 
        (shows ''initial node of ART does not correspond to initial node of program'');
      check_allm (\<lambda> a. case art_edge A a of Cover _ \<Rightarrow> succeed
      | Children tas \<Rightarrow> let l = node_location A a; tau_l' = trans_P_list l
          in check_allm (\<lambda> (\<tau>,l'). check (\<exists> (\<tau>',b) \<in> set tas. \<tau>' = \<tau> \<and> node_location A b = l'
            \<and> (b \<in> art_nodes A)) 
            (shows ''could not find matching transition in ART for node '' o shows a o shows '' and transition '' o shows \<tau>)) 
            tau_l'
      ) ans
    } <+? (\<lambda> s. shows ''could not ensure simulation condition of ART\<newline>'' o s)"

definition error_cond_checker :: "shows check" where
  "error_cond_checker \<equiv> do {
     check_allm (\<lambda> a. if node_location A a = err 
       then case error_hints a of 
         None \<Rightarrow> error (shows ''could not find hint for checking error condition of node '' o shows a)
       | Some hint \<Rightarrow> entailment_checker hint (node_invariant A a) form_False
       else succeed) ans
   } <+? (\<lambda> s. shows ''could not ensure error condition of ART\<newline>'' o s)" 

definition cover_edges_cond_checker :: "shows check" where
  "cover_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a. case art_edge A a of Children _ \<Rightarrow> succeed
       | Cover b \<Rightarrow> do {
           check (b \<in> art_nodes A) (shows ''target node is not listed as art-node'');
           check (node_location A a = node_location A b) (shows ''node-locations differ'');
           check (\<not> is_cover_node (art_edge A b)) (shows ''target node must not have cover edge'');
           case cover_hints a of None \<Rightarrow> error (shows ''there is no cover hint present'')
             | Some hint \<Rightarrow> entailment_checker hint (node_invariant A a) (node_invariant A b)
       } <+? (\<lambda> s. shows ''problem in checking cover edge of node '' o shows a o shows_nl o s)
     ) ans
   } <+? (\<lambda> s. shows ''could not ensure cover edge condition of ART\<newline>'' o s)" 

definition initial_cond_checker :: "shows check" where
  "initial_cond_checker \<equiv> 
     check (node_invariant A (initial_node A) = form_True) 
       (shows ''the node invariant for the initial ART node ('' 
       o (shows (initial_node A)) o shows '') must be TRUE'')" 

definition children_edges_cond_checker :: "shows check" where
  "children_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a. case art_edge A a of Cover _ \<Rightarrow> succeed
       | Children tas \<Rightarrow> do {
           case transition_hints a of None \<Rightarrow> error (shows ''there is no transition hint present'')
             | Some hints \<Rightarrow> do {
               check (length hints = length tas) (shows ''the number of hints differs from the number of children'');
               check_allm (\<lambda> ((\<tau>,b),hint). entailment_checker hint 
                 ((map_formula Pre (node_invariant A a)) and trans_formula_of P \<tau>) 
                  (map_formula Post (node_invariant A b))
                 <+? (\<lambda> s. shows ''problem in checking transition '' o shows a o shows '' --'' o shows \<tau> o shows ''--> ''
                   o shows b o shows_nl o s)) (zip tas hints)
             } 
       } <+? (\<lambda> s. shows ''problem in checking transitions of node '' o shows a o shows_nl o s)
     ) ans
   } <+? (\<lambda> s. shows ''could not ensure transition edge condition of ART\<newline>'' o s)" 

definition art_safety_checker_main :: "shows check" where
  "art_safety_checker_main \<equiv> do {
      initial_cond_checker;
      simulation_cond_checker;
      error_cond_checker;
      cover_edges_cond_checker;
      children_edges_cond_checker
   } <+? (\<lambda> s. shows ''could not ensure safety via ART:\<newline>'' o s)"

context
  assumes ans: "set ans = art_nodes A"
  and trans_P_list: "\<And> l. set (trans_P_list l) = 
    {(\<tau>,trans_target \<tau>') | \<tau> \<tau>'. \<tau> \<in> transitions P \<and> \<tau>' = transition_of P \<tau>  \<and> trans_source \<tau>' = l}"
begin

lemma initial_cond_checker[iff]: "isOK(initial_cond_checker) = initial_cond"
  unfolding initial_cond_def initial_cond_checker_def by auto

lemma error_cond_checker[iff]: "isOK(error_cond_checker) \<Longrightarrow> error_cond"
  unfolding error_cond_def error_cond_checker_def
  by (auto simp: ans split: option.splits dest: entailment_checker)

lemma cover_edges_cond_checker[iff]: "isOK(cover_edges_cond_checker) \<Longrightarrow> cover_edges_cond"
  unfolding cover_edges_cond_def cover_edges_cond_checker_def
  by (auto simp: ans split: art_edge.splits option.splits dest: entailment_checker)

lemma simulation_cond_checker[intro]: assumes "isOK(simulation_cond_checker)"
  shows simulation_cond
proof -
  note ok = assms[unfolded simulation_cond_checker_def, simplified, unfolded Let_def ans]
  show ?thesis
    unfolding simulation_cond_def
  proof (intro conjI allI impI)
    fix a tas l \<tau> \<phi> l'
    assume *: "a \<in> art_nodes A" "art_edge A a = Children tas" "l = node_location A a"
       "\<tau> \<in> transitions P" "transition_of P \<tau> = Transition l \<phi> l'"
    from conjunct2[OF conjunct2[OF ok], rule_format, OF *(1)] *(2-)
    show "\<exists>b\<in>art_nodes A. (\<tau>, b) \<in> set tas \<and> node_location A b = l'" by (fastforce simp: trans_P_list)
  qed (insert ok, auto)
qed

lemma children_edges_cond_checker[intro]: assumes "isOK(children_edges_cond_checker)"
  shows "children_edges_cond"
  unfolding children_edges_cond_def 
proof (intro allI impI)
  fix a tas \<tau> b
  assume a: "a \<in> art_nodes A" 
    and edge: "art_edge A a = Children tas" and mem: "(\<tau>, b) \<in> set tas"
  note ok = assms[unfolded children_edges_cond_checker_def, simplified, unfolded ans, rule_format, OF a, 
    unfolded edge, simplified]
  from ok obtain hints where tr: "transition_hints a = Some hints" and len: "length hints = length tas"
    by (auto split: option.splits)
  from mem len obtain h where mem: "((\<tau>,b),h) \<in> set (zip tas hints)" unfolding set_zip
    by (auto simp: set_conv_nth)
  with ok[unfolded tr, simplified] 
  show "transition_entailment P (node_invariant A a) \<tau> (node_invariant A b)"
    unfolding transition_entailment_def
    by (auto intro!: entailment_checker)
qed

lemma art_safety_checker_main: "isOK(art_safety_checker_main) \<Longrightarrow> safety (Safety_Prog err P)"
  by (intro well_formed_sound, auto simp: art_safety_checker_main_def well_formed_def)
end
end
end

(* Implementation type via lists *)

type_synonym ('v,'l,'t)transitions_impl = "('t \<times> ('v,'l)transition) list"

datatype ('v,'l,'t)lin_prog_impl = Lin_Prog_Impl 'l "('v,'l,'t)transitions_impl"

fun lin_prog_of :: "('v,'l,'t)lin_prog_impl \<Rightarrow> ('v,'l,'t)lin_prog" where
  "lin_prog_of (Lin_Prog_Impl i ts) = (Lin_Prog i (set (map fst ts)) (the o map_of ts))"

(* TODO: improve efficiency! *)
fun succ_transitions :: "('v,'l,'t)lin_prog_impl \<Rightarrow> ('v,'l,'t)lin_prog \<Rightarrow> 'l \<Rightarrow> ('t \<times> 'l)list" where
  "succ_transitions (Lin_Prog_Impl i ts) (Lin_Prog _ _ \<tau>s) l = 
    [ (\<tau>,l') . (\<tau>,_) \<leftarrow> ts, Transition ll \<phi> l' \<leftarrow> [\<tau>s \<tau>], ll = l]"

lemma succ_transitions: "set (succ_transitions P (lin_prog_of P) l) = 
  {(\<tau>, trans_target \<tau>') |\<tau> \<tau>'.
    \<tau> \<in> transitions (lin_prog_of P) \<and> \<tau>' = transition_of (lin_prog_of P) \<tau> \<and> trans_source \<tau>' = l}"
  by (cases P, auto split: transition.splits; force?, (metis transition.sel transition.collapse)?)

datatype 't art_edge_impl = 
  Cover_Edge art_node entailment_info 
| Children_Edge "('t \<times> art_node \<times> entailment_info)list"

datatype ('v,'l,'t)art_node_impl = Art_Node 
  (art_node: art_node) (invariant: "'v invariant") (location: 'l) (art_edge: "'t art_edge_impl")

fun is_cover_node :: "('v,'l,'t)art_node_impl \<Rightarrow> bool" where
  "is_cover_node (Art_Node _ _ _ (Cover_Edge _ _)) = True"
| "is_cover_node (Art_Node _ _ _ (Children_Edge _)) = False"

record ('v,'l,'t)art_impl = 
  initial_node :: art_node
  art_nodes :: "('v,'l,'t)art_node_impl list"
  error_hints :: "(art_node \<times> entailment_info)list"

definition art_nodes :: "('v,'l,'t)art_impl \<Rightarrow> art_node list" where 
  "art_nodes A = (map art_node (art_impl.art_nodes A))"

fun art_edge_of :: "'t art_edge_impl \<Rightarrow> 't art_edge" where
  "art_edge_of (Cover_Edge an _) = Cover an"
| "art_edge_of (Children_Edge ans) = Children (map (\<lambda> (t,a,_). (t,a)) ans)"

definition art_of :: "('v,'l,'t)art_impl \<Rightarrow> ('v,'l,'t)art" where
  "art_of A \<equiv> let i = art_impl.initial_node A; ans = art_impl.art_nodes A in 
    \<lparr> 
      art.initial_node = i,
      art_nodes = set (art_nodes A),
      art_edge = (the o map_of (map (\<lambda> a. (art_node a, art_edge_of (art_edge a))) ans)),
      node_location = (the o map_of (map (\<lambda> a. (art_node a, location a)) ans)),
      node_invariant = (the o map_of (map (\<lambda> a. (art_node a, invariant a)) ans))
    \<rparr>" 

definition error_hints :: "('v,'l,'t)art_impl \<Rightarrow> art_node \<Rightarrow> entailment_info option" where
  "error_hints A = map_of (art_impl.error_hints A)"

definition cover_hints :: "('v,'l,'t)art_impl \<Rightarrow> art_node \<Rightarrow> entailment_info option" where
  "cover_hints A = map_of (map (\<lambda> a. (art_node a, 
      case art_edge a of Cover_Edge _ h \<Rightarrow> h)) 
    (filter is_cover_node (art_impl.art_nodes A)))"

definition transition_hints :: "('v,'l,'t)art_impl \<Rightarrow> art_node \<Rightarrow> entailment_info list option" where
  "transition_hints A = map_of (map (\<lambda> a. (art_node a, 
      case art_edge a of Children_Edge cs \<Rightarrow> map (\<lambda> (_,_,h). h) cs)) 
    (filter (\<lambda> a. \<not> is_cover_node a) (art_impl.art_nodes A)))"

definition art_safety_checker :: "('v :: {compare_order,show},'l :: show,'t :: show)lin_prog_impl \<Rightarrow> 'l 
  \<Rightarrow> ('v,'l,'t)art_impl \<Rightarrow> shows check" where
  "art_safety_checker Pi err Ai \<equiv> let
     P = lin_prog_of Pi;
     A = art_of Ai;
     checker = art_safety_checker_main err P A 
       (art_nodes Ai) 
       (succ_transitions Pi P)
       (error_hints Ai)
       (cover_hints Ai)
       (transition_hints Ai)
   in
     checker <+? (\<lambda> s. shows ''could not ensure safety w.r.t. error location '' o shows err o shows '' via ART:\<newline>'' o s)"

lemma art_safety_checker: assumes "isOK(art_safety_checker Pi err Ai)"
  shows "safety (Safety_Prog err (lin_prog_of Pi))"
proof -
  have ok: "isOK (art_safety_checker_main err (lin_prog_of Pi) (art_of Ai)
         (art_nodes Ai) (succ_transitions Pi (lin_prog_of Pi))
         (error_hints Ai) (cover_hints Ai) (transition_hints Ai))"
    using assms[unfolded art_safety_checker_def Let_def] by simp
  show ?thesis
    by (rule art_safety_checker_main[OF _ succ_transitions ok], auto simp: art_nodes_def art_of_def Let_def)
qed

end
