(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)
theory Exact_Tree_Automata_Completion_Impl
imports
  Tree_Automata_Impl
  Exact_Tree_Automata_Completion
  "../Termination_and_Complexity/Signature_Extension"
begin

lemma states_map_states: "inj f \<Longrightarrow> states (map_states f A) = f ` states A"
using states_map_states_ex by fastforce

lemma fground_rsteps:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes "(a, 0) \<in> F"
    and rsteps: "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
    and "funas_trs R \<subseteq> F" and "funas_term s \<subseteq> F" and "funas_term t \<subseteq> F"
  shows "\<exists>\<sigma> \<tau>. fground F (s \<cdot> \<sigma>) \<and> fground F (t \<cdot> \<tau>) \<and> (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof -
  interpret cleaning_const F "\<lambda>_. Fun a []" "Fun a []"
    using assms by (unfold_locales) (auto)
  let ?\<sigma> = "star a \<circ> clean_subst \<sigma>"
  let ?\<tau> = "star a \<circ> clean_subst \<tau>"
  from rsteps_imp_clean_rsteps [OF \<open>funas_trs R \<subseteq> F\<close> rsteps]
    have "(clean_term (s \<cdot> \<sigma>), clean_term (t \<cdot> \<tau>)) \<in> (rstep R)\<^sup>*" .
  then have *: "(s \<cdot> clean_subst \<sigma>, t \<cdot> clean_subst \<tau>) \<in> (rstep R)\<^sup>*"
    using \<open>funas_term s \<subseteq> F\<close> and \<open>funas_term t \<subseteq> F\<close> by (auto simp: sig_step_def)
  from star_rsteps [OF this] have "(s \<cdot> ?\<sigma>, t \<cdot> ?\<tau>) \<in> (rstep R)\<^sup>*" by simp
  moreover have "fground F (s \<cdot> ?\<sigma>)" and "fground F (t \<cdot> ?\<tau>)"
    using \<open>(a, 0) \<in> F\<close> and \<open>funas_term s \<subseteq> F\<close> and \<open>funas_term t \<subseteq> F\<close>
    by (auto simp: funas_term_star_conv funas_term_subst split: if_splits)
  ultimately show ?thesis by blast
qed

lemma set_ta_rules_impl' [simp]: "set (ta_rules_impl' A) = ta_rules (ta_of_ta A)"
by (cases A) auto

lemma (in etac) [simp]: "ta_eps mp_ta = {}" by (auto simp: mp_ta_def)

lemma (in etac) [simp]: "ta_final mp_ta = {}" by (auto simp: mp_ta_def)

lemma size_list_cong [cong]:
  "(\<And>x. x \<in> set xs \<Longrightarrow> f x = g x) \<Longrightarrow> size_list f xs = size_list g xs"
by (induct xs) force+

lemma size_star [simp]:
  "(size \<circ>\<circ> star) c = size"
by (rule ext, induct_tac x) (auto)

lemma size_r_rhs_le:
  assumes "r \<in> ground_instances_rules F c t"
  shows "size (r_rhs r) \<le> size t"
using assms
apply (induct t)
apply (auto simp: sig_rules_def)
by (simp add: le_Suc_eq size_list_estimation')

lemma ta_rules_ground_instances_ta_unique:
  assumes "TA_rule g qs (star c (Fun f ts)) \<in> ta_rules (ground_instances_ta F c (Fun f ts))"
  shows "g = f \<and> qs = map (star c) ts"
proof -
  let ?t = "Fun f ts"
  have "TA_rule g qs (star c ?t) \<in> ground_instances_rules F c ?t"
    using assms by (force simp: ground_instances_ta_def)
  then have "TA_rule g qs (star c ?t) \<in> {TA_rule f (map (star c) ts) (star c ?t)} \<union>
    \<Union>(ground_instances_rules F c ` set ts)"
    by force
  moreover have "TA_rule g qs (star c ?t) \<notin> \<Union>(ground_instances_rules F c ` set ts)"
    apply (auto dest!: size_r_rhs_le)
    using not_le size_simp1 by auto
  ultimately show ?thesis by auto
qed

lemma star_Fun_in_reachable_states_ground_instances_ta_imp_Fun:
  assumes "star c (Fun f ts) \<in> reachable_states (ground_instances_ta F c (Fun f ts)) (fterm s)"
  shows "\<exists>ss. s = Fun f ss"
proof -
  let ?A = "\<lambda>t. ground_instances_ta F c t"
  let ?t = "Fun f ts"
  show "\<exists>ss. s = Fun f ss"
  proof (cases s)
    case (Var x)
    show ?thesis using assms [unfolded Var] by simp
  next
    case (Fun g ss)
    let ?ss = "map fterm ss"
    let ?qs = "map (star c) ss"
    have "fterm (Fun g ss) = (Fun (Inl g) ?ss)" by simp
    with assms Fun have "star c ?t \<in> reachable_states (?A ?t) (Fun (Inl g) ?ss)" by simp
    then obtain q qs
      where rule: "TA_rule g qs q \<in> ta_rules (?A ?t)"
      and eps: "(q, star c ?t) \<in> (ta_eps (?A ?t))\<^sup>*"
      and "length qs = length ?ss"
      and "\<forall>i < length ?ss. qs ! i \<in> reachable_states (?A ?t) (?ss ! i)" by auto
    have "q = star c ?t" using eps ground_instances_ta_def [of F c ?t] by auto
    then have "TA_rule g qs (star c ?t) \<in> ta_rules (?A ?t)" using rule by auto
    from ta_rules_ground_instances_ta_unique [OF this]
      have "g = f \<and> qs = map (star c) ts" .
    then have "s = Fun f ss" using Fun by blast
    then show ?thesis by simp
  qed
qed

lemma fground_reachable_states_ground_instances_ta:
  assumes "fground F t"
  shows "Fun c [] \<in> reachable_states (ground_instances_ta F c (Var x)) (fterm t)"
  (is "?c \<in> reachable_states ?A ?t")
using assms
proof (induct t)
  case (Fun f ts)
  let ?ts = "map fterm ts"
  thm reachable_states.simps(2)
  have "(f, length ts) \<in> F" using Fun by force
  then have 1: "TA_rule f (replicate (length ts) ?c) ?c \<in> ta_rules ?A"
    by (auto simp: ground_instances_ta_def sig_rules_def)
  have 2: "(?c, ?c) \<in> (ta_eps ?A)\<^sup>*" by auto
  have 3: "length (replicate (length ts) (Fun c [])) = length ?ts" by auto
  { fix i
    assume "i < length ts"
    then have "ts ! i \<in> set ts" by auto
    moreover then have "fground F (ts ! i)" using Fun by auto
    ultimately have "?c \<in> reachable_states ?A (fterm (ts ! i))" using Fun by auto
  }
  then have 4: "\<forall>i < length ts. ?c \<in> reachable_states ?A (fterm (ts ! i))" by auto
  show ?case using reachable_states_InlI [OF 1 2 3] 4 by force
qed (auto)

lemma ta_rules_ground_instances_ta:
  assumes "s \<unrhd> t"
  shows "ta_rules (ground_instances_ta F c s) = ground_instances_rules F c t \<union>
   (ground_instances_rules F c s - ground_instances_rules F c t)"
using assms by (induct t) (auto simp: ground_instances_ta_def)

lemma reachable_states_mono:
  assumes "ta_rules A \<subseteq> ta_rules B"
    and "ta_eps A \<subseteq> ta_eps B"
  shows "reachable_states A t \<subseteq> reachable_states B t"
proof
  fix q
  assume "q \<in> reachable_states A t"
  then have "(t, State q) \<in> (move A)\<^sup>*" using move_reachable_states [symmetric] by fast
  from move_mono'_rtrancl [OF this assms] have "(t, State q) \<in> (move B)\<^sup>*" .
  then show "q \<in> reachable_states B t" using move_reachable_states by fast
qed

lemma ground_instances_subseteq_lang_ground_instances_ta:
  assumes "funas_term t \<subseteq> F"
  shows "ground_instances F t \<subseteq> lang (ground_instances_ta F c t)" 
  (is "?\<Sigma> t \<subseteq> lang (?A t)")
using assms
proof (induct t)
  case (Var x)
  then show ?case by simp
next
  case (Fun f ts)
  then have IH: "\<And>i. i < length ts \<Longrightarrow> ?\<Sigma> (ts ! i) \<subseteq> lang (?A (ts ! i))"
    by (auto dest: nth_mem)
  { fix ss
    let ?rs = "\<lambda>A. (\<lambda>t. reachable_states A t)"
    assume l: "length ss = length ts"
      and ss: "\<forall>i < length ss. ss ! i \<in> ?\<Sigma> (ts ! i)"
    with IH have *: "\<forall>i < length ss. star c (ts ! i) \<in> ?rs (?A (ts ! i)) (fterm (ss ! i))"
      by (auto simp: lang_def)
    { fix i
      assume len: "i < length ss"
      then have s: "(Fun f ts) \<unrhd> ts ! i" by (auto simp: l)
      have 1: "ta_rules (?A (ts ! i)) \<subseteq> ta_rules (?A (Fun f ts))"
        using ground_instances_rules_mono [OF s] by (auto simp: ground_instances_ta_def)
      have 2: "ta_eps (?A (ts ! i)) \<subseteq> ta_eps (?A (Fun f ts))"
        by (auto simp: ground_instances_ta_def)
      have 3: "star c (ts ! i) \<in> ?rs (?A (ts ! i)) (fterm (ss ! i))"
        using * len by auto
      from reachable_states_mono [OF 1 2] 3
        have "star c (ts ! i) \<in> ?rs (?A (Fun f ts)) (fterm (ss ! i))" by blast
    }
    then have "\<forall>i < length ss. star c (ts ! i) \<in> ?rs (?A (Fun f ts)) (fterm (ss ! i))"
      by auto
    moreover have "TA_rule f (map (star c) ts) (star c (Fun f ts)) \<in> ta_rules (?A (Fun f ts))"
      by (auto simp: ground_instances_ta_def)
    ultimately have "star c (Fun f ts) \<in> ?rs (?A (Fun f ts)) (fterm (Fun f ss))"
      by (auto intro!: exI [of _ "star c (Fun f ts)"] exI [of _ "map (star c) ts"] simp: l)
    moreover have "ground (Fun f ss)"
      using ss by (auto simp: ground_instances_def) (metis ground_subst in_set_conv_nth)
    moreover have "{star c (Fun f ts)} = ta_final (?A (Fun f ts))" by auto
    ultimately have "Fun f ss \<in> lang (?A (Fun f ts))" by (auto simp: lang_def)
  }
  then show ?case using term_ground_instances_Fun [of F f ts] by (auto)
qed

(* A version of reachable_states which works on sets of rules and is equivalent to
   the other version if there are no epsilon-transitions. *)
fun reachable_states' :: "('q, 'f) ta_rule set \<Rightarrow> ('f + 'q, 'v) term \<Rightarrow> 'q set"
where
  "reachable_states' \<Delta> (State q) = {q}"
| "reachable_states' \<Delta> (Fun (Inl f) ts) =
    r_rhs ` {r \<in> \<Delta>. r_root r = f \<and> length (r_lhs_states r) = length ts \<and>
      (\<forall>i < length ts. (r_lhs_states r) ! i \<in> reachable_states' \<Delta> (ts ! i))}"
| "reachable_states' \<Delta> _ = {}"

export_code reachable_states' in Haskell

lemma reachable_states_reachable_states':
  assumes "ta_eps A = {}"
  shows "reachable_states A t = reachable_states' (ta_rules A) t"
using assms
proof (induct t rule: reachable_states.induct)
  case (2 A f ts)
  then show ?case
  apply auto
  defer
  apply (metis Tree_Automata.ta_rule.collapse)
  proof -
    fix x :: 'a and qs :: "'a list"
    assume a1: "\<forall>i<length ts. qs ! i \<in> reachable_states' (ta_rules A) (ts ! i)"
    assume a2: "f qs \<rightarrow> x \<in> ta_rules A"
    assume a3: "length qs = length ts"
    have f4: "\<forall>p t. \<not> p (t::('a, 'b) ta_rule) \<or> t \<in> Collect p"
      by (meson CollectI)
    have "\<forall>a f t T. ((a::'a) \<noteq> f (t::('a, 'b) ta_rule) \<or> t \<notin> T) \<or> a \<in> f ` T"
      by blast
    then show "x \<in> r_rhs ` {t \<in> ta_rules A. r_root t = f \<and> length (r_lhs_states t) = length ts \<and>
      (\<forall>n < length ts. r_lhs_states t ! n \<in> reachable_states' (ta_rules A) (ts ! n))}"
      using f4 a3 a2 a1 by force
  qed
qed (auto)

lemma reachable_states_reachable_states'_iff:
  assumes "ta_eps A = {}"
  shows "q \<in> reachable_states A t \<longleftrightarrow> q \<in> reachable_states' (ta_rules A) t"
using reachable_states_reachable_states' [OF assms] by (intro iffI) blast+

lemma reachable_states'_mono:
  assumes "\<Delta> \<subseteq> \<Delta>'"
  shows "reachable_states' \<Delta> t \<subseteq> reachable_states' \<Delta>' t"
using assms by (induct t rule: reachable_states'.induct, auto) blast

(* a list of all assignments from 'a's to 'b's *)
fun combs :: "'a list \<Rightarrow> 'b list \<Rightarrow> ('a \<times> 'b) list list"
where
  "combs [] ys = [[]]"
| "combs (x # xs) ys = concat (map (\<lambda>l. (map (\<lambda>y. (x, y) # l) ys)) (combs xs ys))"

value "combs [a,b,c] [x,y,z]"

definition state_substs :: "'v list \<Rightarrow> 'q list \<Rightarrow> ('v \<times> 'q) list list" where
  "state_substs V Q \<equiv> combs V Q"

export_code state_substs in Haskell

fun qi' :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('v \<Rightarrow> 'q + ('f, 'v) term) \<Rightarrow> ('f, 'v) term \<Rightarrow> 'q + ('f, 'v) term"
where
  "qi' c t g (Var x) = (if x \<in> vars_term t then g x else Inr (star c (Var x)))"
| "qi' c t g (Fun f ts) = Inr (star c (Fun f ts))"

definition inf_step :: "'f \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('v \<times> ('q + ('f, 'v) term)) list list
  \<Rightarrow> ('q + ('f, 'v) term, 'f) ta_rule set \<Rightarrow> ('q + ('f, 'v) term, 'f) ta_rule set"
where
  "inf_step c R S \<Delta> = \<Union>(set (map (\<lambda>((l, r), \<theta>). (case l of Fun f ls \<Rightarrow>
    (\<lambda>q. TA_rule f (map (\<lambda>li. qi' c r (fun_of \<theta>) li) ls) q) `
    (reachable_states' \<Delta> (fterm r \<cdot> (State \<circ> (fun_of \<theta>)) :: ('f, 'q, 'v) crterm))))
    (List.product R S))
  )"

lemma (in etac) qi_qi' [simp]:
  "qi' c t g s = qi t g s"
by (cases s) auto

lemma (in etac) qi_qi'_2:
  assumes "\<forall>x \<in> vars_term t. g x = g' x"
  shows "qi' c t g' s = qi t g s"
using assms by (cases s) auto

lemma (in etac) qi_qi'_3:
  assumes "\<forall>x \<in> vars_term t. g x = g' x"
  shows "qi' c t g' = qi t g"
using qi_qi'_2 [OF assms] by blast


lemma inf_stepE:
  assumes "\<forall>(l, r) \<in> set R. is_Fun l"
    and  "TA_rule f qs q \<in> inf_step c R S \<Delta>"
  obtains ls r \<theta> where "(Fun f ls, r) \<in> set R" and "\<theta> \<in> set S"
    and "q \<in> reachable_states' \<Delta> (fterm r \<cdot> (State \<circ> fun_of \<theta>) ::
      ('b + 'a + ('b, 'c) term, 'c) term)"
    and "qs = map (qi' c r (fun_of \<theta>)) ls"
  using assms by (fastforce simp: inf_step_def)

lemma inf_stepI:
  assumes "(Fun f ls, r) \<in> set R" and "\<theta> \<in> set S"
    and "q \<in> reachable_states' \<Delta> (fterm r \<cdot> (State \<circ> fun_of \<theta>) ::
      ('b + 'a + ('b, 'c) term, 'c) term)"
    and "qs = map (qi' c r (fun_of \<theta>)) ls"
  shows "TA_rule f qs q \<in> inf_step c R S \<Delta>"
using assms by (auto simp: inf_step_def split: prod.splits term.splits)

lemma inf_step_mono:
  assumes "\<forall>(l, r) \<in> set R. is_Fun l" and "\<Delta> \<subseteq> \<Delta>'"
  shows "inf_step c R S \<Delta> \<subseteq> inf_step c R S \<Delta>'" (is "?L \<subseteq> ?R")
proof
  fix tr
  assume "tr \<in> ?L"
  moreover then obtain f qs q where [simp]: "TA_rule f qs q = tr"
    by (metis Tree_Automata.r_sym.cases)
  ultimately obtain ls r \<theta> where 1: "(Fun f ls, r) \<in> set R" and 2: "\<theta> \<in> set S"
    and rs: "q \<in> reachable_states' \<Delta> (fterm r \<cdot> (State \<circ> fun_of \<theta>) ::
      ('a + 'c + ('a, 'b) term, 'b) term)"
    and 3: "qs = map (qi' c r (fun_of \<theta>)) ls"
    using inf_stepE [OF assms(1)] by blast
  have "q \<in> reachable_states' \<Delta>' (fterm r \<cdot> (State \<circ> fun_of \<theta>) ::
      ('a + 'c + ('a, 'b) term, 'b) term)"
    using reachable_states'_mono [OF assms(2)] rs by auto
  from inf_stepI [OF 1 2 this 3] have "TA_rule f qs q \<in> inf_step c R S \<Delta>'" .
  then show "tr \<in> ?R" by auto
qed

abbreviation "r_vars R \<equiv> remdups (concat (map (vars_term_list \<circ> snd) R))"

definition graph_of :: "'v list \<Rightarrow> ('v \<Rightarrow> 'q) \<Rightarrow> ('v \<times> 'q) list"
where
  "graph_of V f = map (\<lambda>x. (x, f x)) V"

lemma fun_of_graph_of:
  assumes "x \<in> set V"
  shows "fun_of (graph_of V f) x = f x"
using assms by (induct V) (auto simp: graph_of_def fun_of_def)

lemma combs_graph_of:
  assumes "distinct V"
  shows "set (combs V Q) = {graph_of V f|f. f ` set V \<subseteq> set Q}"
using assms
proof (induct V)
  case (Cons v vs)
  then have "set (combs vs Q) = {graph_of vs f |f. f ` set vs \<subseteq> set Q}" by force
  moreover {
    fix y f
    assume "y \<in> set Q"
      and "f ` set vs \<subseteq> set Q"
    with Cons(2) have "\<exists>g. y = g v \<and> (\<forall>x\<in>set vs. f x = g x) \<and> g v \<in> set Q \<and> g ` set vs \<subseteq> set Q"
    by (intro exI [of _ "\<lambda>x. (if x = v then y else f x)"]) auto
  }
  ultimately show ?case unfolding combs.simps using Cons(2)
  by (simp only: set_concat set_map) (auto simp: graph_of_def)
qed (auto simp: graph_of_def)

lemma ta_eps_ground_instances_ta [simp]:
  "ta_eps (ground_instances_ta F c t) = {}"
by (induct t) (auto simp: ground_instances_ta_def)

lemma states_ground_instances_ta_Var [simp]:
  "states (ground_instances_ta F c (Var x)) = {Fun c []}"
by (auto simp: ground_instances_ta_def sig_rules_def states_def r_states_def)

lemma states_ground_instances_ta_Fun [simp]:
  "states (ground_instances_ta F c (Fun f ts)) =
   {star c (Fun f ts)} \<union> \<Union>(states ` (ground_instances_ta F c) ` set ts)"
by (auto simp: ground_instances_ta_def sig_rules_def states_def r_states_def)

definition "sr_impl R = remdups (concat (map (supt_list \<circ> fst) R))"

lemma sr_impl [simp]: "set (sr_impl R) = sr (set R)"
by (induct R) (auto simp: sr_def sr_impl_def)

definition "sig_rules_list F c = map (\<lambda>(f, n). TA_rule f (replicate n c) c) F"

lemma sig_rules_list [simp]: "set (sig_rules_list F c) = sig_rules (set F) c"
by (auto simp: sig_rules_list_def sig_rules_def)

fun gi_rules_list :: "('f \<times> nat) list \<Rightarrow> 'f \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f ,'v) term, 'f) ta_rule list"
where
  "gi_rules_list F c (Var x) = sig_rules_list F (Fun c [])"
| "gi_rules_list F c (Fun f ts) =
    TA_rule f (map (star c) ts) (star c (Fun f ts)) # concat (map (gi_rules_list F c) ts)"

lemma gi_rules_list [simp]:
  "set (gi_rules_list F c t) = ground_instances_rules (set F) c t"
by (induct t) (auto)

definition "mp_ta_rules R F c =
  sig_rules_list F (Fun c []) @
  concat (map (\<lambda>t. case t of
    Var _ \<Rightarrow> []
  | Fun f ts \<Rightarrow> [TA_rule f (map (star c) ts) (star c t)]) (sr_impl R))"

lemma (in etac) mp_ta_rules [simp]:
  assumes "set R' = R" and "set F' = F"
  shows "set (mp_ta_rules R' F' c) = ta_rules (mp_ta)"
using assms by (auto simp: mp_ta_rules_def mp_ta_def)

lemma states_ground_instances_ta_subseteq:
  "states (ground_instances_ta F c t) \<subseteq> {Fun c []} \<union> {star c s | s. s \<unlhd> t}"
by (induct t) (simp_all, force)

lemma states_ground_instances_ta_ground:
  assumes "ground t"
  shows "states (ground_instances_ta F c t) \<subseteq> {star c s | s. s \<unlhd> t}"
using assms by (induct t) (simp_all, force)

lemma not_ground_star_in_reachable_states:
  assumes "\<not> ground t" and "(a, 0) \<in> F"
  shows "Fun c [] \<in> reachable_states (ground_instances_ta F c t) (fterm (Fun a []))"
using assms by (induct t) (auto simp: ground_instances_ta_def sig_rules_def)

lemma star_ground_instances:
  assumes "(a, 0) \<in> F" and "funas_term t \<subseteq> F"
  shows "star a t \<in> ground_instances F t"
using assms by (induct t)
  (force simp: ground_instances_def star_subst_conv UN_subset_iff intro: exI [of _ "\<lambda>x. Fun a []"])+

lemma star_reachable_states_ground_instances_ta_star:
  assumes "(a, 0) \<in> F" and "funas_term t \<subseteq> F"
  shows "star c t \<in> reachable_states (ground_instances_ta F c t) (fterm (star a t))"
using star_ground_instances [OF assms]
  and ground_instances_subseteq_lang_ground_instances_ta [OF assms(2)]
  by (auto simp: lang_def)

lemma subterm_star_reachable_states_ground_instances_ta:
  assumes "(a, 0) \<in> F" and "funas_term t \<subseteq> F"
    and "s \<unlhd> t"
  shows "star c s \<in> reachable_states (ground_instances_ta F c t) (fterm (star a s))"
using ground_instances_rules_mono [OF assms(3), of F c]
  and star_reachable_states_ground_instances_ta_star [of a F s c]
  and supteq_imp_funas_term_subset [OF assms(3)]
  and assms(1,2)
by (auto simp: ground_instances_ta_def reachable_states_reachable_states'
         dest: reachable_states'_mono)  

lemma const_fground:
  assumes "(a, 0) \<in> F"
  shows "fground F (Fun a [])"
using assms by auto

lemma accessible_ground_instances_ta:
  fixes t :: "('f, 'v) term"
  assumes "(a, 0) \<in> F" and "funas_term t \<subseteq> F"
  shows "\<forall>q \<in> states (ground_instances_ta F c t). accessible TYPE('v) (ground_instances_ta F c t) q"
    (is "\<forall>q \<in> ?Q. ?acc q")
proof -
  { fix s
    assume "s \<unlhd> t"
    then have "?acc (star c s :: ('f, 'v) term)"
      using subterm_star_reachable_states_ground_instances_ta [of a F t s c] and assms
      by (force simp: accessible_def)
  }
  note * = this
  show ?thesis
  proof (cases "ground t")
    case (True)
    then have "?Q \<subseteq> {star c s | s. t \<unrhd> s}" by (rule states_ground_instances_ta_ground)
    with * show ?thesis by blast
  next
    case (False)
    then have "?acc (Fun c [])" using not_ground_star_in_reachable_states [of t a F c] assms(1)
      by (auto dest: const_fground simp: accessible_def)
    moreover have "?Q \<subseteq> {Fun c []} \<union> {star c s | s. t \<unrhd> s}"
      by (rule states_ground_instances_ta_subseteq)
    ultimately show ?thesis using * by blast
  qed
qed

locale gi_etac =
  etac c F R a A for c F and R :: "('f, 'v) trs" and a and A :: "('q, 'f) ta" +
  assumes eps_empty: "ta_eps A = {}"
begin

lemma overapproximation_of_\<Delta>\<^sub>C:
  assumes "ta_rules cr_ta \<subseteq> \<Delta>"
    and "set R' = R"
    and "set Q = states cr_ta"
    and "inf_step c R' (state_substs (r_vars R') Q) \<Delta> \<subseteq> \<Delta>" (is "inf_step _ _ ?S _ \<subseteq> _")
  shows "\<Delta>\<^sub>C \<subseteq> \<Delta>"
proof (intro subsetI)
  fix tr
  assume "tr \<in> \<Delta>\<^sub>C"
  then show "tr \<in> \<Delta>"
  proof (induct)
    case (base r)
    then show ?case using assms by auto
  next
    case (step f ls r \<theta> q)
    let ?A = "cr_ta\<lparr>ta_rules := {x \<in> \<Delta>\<^sub>C. x \<in> \<Delta>}\<rparr>"
    have *: "q \<in> reachable_states ?A (fterm r \<cdot> (State \<circ> \<theta>) :: ('f, 'q, 'v) crterm)"
      using move_reachable_states [THEN iffD1, OF step(4)] .
    have "ta_eps ?A = {}" using eps_empty by (auto simp: cr_ta_def mp_ta_def map_states_def)
    from reachable_states_reachable_states'_iff [OF this, THEN iffD1, OF *]
      have 3: "q \<in> reachable_states' {x \<in> \<Delta>\<^sub>C. x \<in> \<Delta>} (fterm r \<cdot> (State \<circ> \<theta>) :: ('f, 'q, 'v) crterm)" by auto
    have **: "range \<theta> \<subseteq> set Q" using step assms by simp
    let ?V = "r_vars R'"
    let ?\<theta> = "graph_of ?V \<theta>"
    have 9: "\<forall>x \<in> set ?V. \<theta> x = fun_of ?\<theta> x" using fun_of_graph_of [of _ ?V \<theta>] by auto
    have 8: "?\<theta> \<in> {graph_of ?V f |f. f ` set ?V \<subseteq> set Q}" using ** by fast
    have "distinct (r_vars R')" by (auto)
    from combs_graph_of [OF this, of Q]
      have "set (combs ?V Q) = {graph_of ?V f |f. f ` set ?V \<subseteq> set Q}" .
    then have 2: "?\<theta> \<in> set (state_substs ?V Q)" using 8 by (auto simp: state_substs_def)
    have 4: "map (qi' c r (fun_of ?\<theta>)) ls = map (qi' c r (fun_of ?\<theta>)) ls" ..
    from 9 have "(fterm r \<cdot> (State \<circ> \<theta>) :: ('f, 'q, 'v) crterm) =
      (fterm r \<cdot> (State \<circ> fun_of ?\<theta>) :: ('f, 'q, 'v) crterm)"
      using assms step by (simp add: term_subst_eq_conv)
    with 3 have 7: "q \<in> reachable_states' {x \<in> \<Delta>\<^sub>C. x \<in> \<Delta>} (fterm r \<cdot> (State \<circ> fun_of ?\<theta>) ::
        ('f, 'q, 'v) crterm)" by simp
    have "(Fun f ls, r) \<in> set R'" using step assms by fast
    from inf_stepI [OF this 2 7 4]
      have 6: "TA_rule f (map (qi' c r (fun_of ?\<theta>)) ls) q \<in> inf_step c R' ?S {x \<in> \<Delta>\<^sub>C. x \<in> \<Delta>}"
      using assms by auto
    have 5: "{x \<in> \<Delta>\<^sub>C. x \<in> \<Delta>} \<subseteq> \<Delta>" by auto
    have "TA_rule f (map (qi' c r (fun_of ?\<theta>)) ls) q \<in> inf_step c R' ?S \<Delta>"
      using inf_step_mono [OF novar [unfolded assms(2) [symmetric]] 5] 6 by auto
    from assms(4) [THEN subsetD, OF this]
      have "TA_rule f (map (qi' c r (fun_of ?\<theta>)) ls) q \<in> \<Delta>" .
    then show ?case using qi_qi'_3 9 \<open>(Fun f ls, r) \<in> set R'\<close> by auto
  qed
qed

end

lemma finite_states_ground_instances_ta [intro]:
  "finite F \<Longrightarrow> finite (states (ground_instances_ta F c t))"
by (induct t) auto

lemma finite_sig_rules [intro]:
  assumes "finite F"
  shows "finite (sig_rules F c)"
proof -
  have [simp]: "sig_rules F c = (\<lambda>(f, n). TA_rule f (replicate n c) c) ` F"
    by (auto simp: sig_rules_def)
  show ?thesis using assms by simp
qed

lemma finite_ta_rules_ground_instances_ta [intro]:
  "finite F \<Longrightarrow> finite (ta_rules (ground_instances_ta F c t))"
by (induct t) (auto simp: ground_instances_ta_def)

definition "check_rules_subseteq rs A =
  check_subseteq rs (ta_rules_impl' A) <+? (\<lambda>r. shows ''rule '' \<circ> shows r \<circ> shows '' is missing'')"

lemma check_rules_subseteq [simp]:
  "isOK (check_rules_subseteq rs A) \<longleftrightarrow> set rs \<subseteq> set (ta_rules_impl' A)"
by (auto simp: check_rules_subseteq_def)

definition "add_rule_states rs ss =
  fold (\<lambda>r ss. case r of TA_rule f qs q \<Rightarrow> List.insert q (fold (List.insert) qs ss)) rs ss"

lemma add_rule_states_Nil [simp]: "add_rule_states [] ss = ss" by (simp add: add_rule_states_def)

lemma add_rule_states_Cons [simp]: "add_rule_states (r # rs) =
  add_rule_states rs \<circ> List.insert (r_rhs r) \<circ> fold List.insert (r_lhs_states r)"
by (cases r) (auto simp: add_rule_states_def [abs_def])

lemma set_add_rule_states [simp]:
  "set (add_rule_states rs ss) = \<Union>(r_states ` set rs) \<union> set ss"
by (induct rs arbitrary: ss) (auto simp: r_states_def List.union_def [symmetric])

definition "check_growing R = check (growing (set R)) (shows ''TRS is not growing'')"

lemma check_growing [simp]:
  "isOK (check_growing R) = growing (set R)"
by (simp add: check_growing_def)

type_synonym ('f, 'v) etac_ta = "(('f, 'v) term + ('f, 'v) term, 'f) tree_automaton"

(*needed due to later container setup (in Proof_Checker/Container_Setup)*)
definition "ta_inter_eps_empty T T' = \<lparr>
  ta_final = ta_final T \<times> ta_final T',
  ta_rules = (\<lambda>r. case r of (TA_rule f ps p, TA_rule g qs q) \<Rightarrow> (TA_rule f (zip ps qs) (p, q))) `
    (\<Union> ((\<lambda> f. {r \<in> ta_rules T. r_sym r = f} \<times> {r \<in> ta_rules T'. r_sym r = f}) ` ta_syms T)),
  ta_eps = {}
\<rparr>"

lemma ta_inter_eps_empty [simp]:
  "ta_eps A = {} \<Longrightarrow> ta_eps B = {} \<Longrightarrow> ta_inter_eps_empty A B = intersect_ta A B"
by (simp add: intersect_ta_def prod_ta_code ta_inter_eps_empty_def)

definition
  check_etac_nonreachable ::
    "('f \<times> nat) list \<Rightarrow> 'f \<Rightarrow> 'f \<Rightarrow> ('f, 'v) etac_ta \<Rightarrow>
      ('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_etac_nonreachable F' a c A' R' s t = do {
    let F = set F';
    check ((a, 0) \<in> F) (shows ''constant '' \<circ> shows a \<circ> shows '' is not in signature'');
    check ((c, 0) \<notin> F) (shows ''star-symbol is not fresh w.r.t. signature'');
    check_subseteq (funas_term_impl s) F' <+? (\<lambda>_. shows ''lhs violates signature'');
    check_subseteq (funas_term_impl t) F' <+? (\<lambda>_. shows ''rhs violates signature'');
    let fs = funas_trs_impl R';
    check_subseteq fs F' <+? (\<lambda>_. shows ''TRS violates signature'');
    check_varcond_no_Var_lhs R';
    check_linear_trs R';
    check_growing R';
    let A = ta_of_ta A';
    let q = Inl (star c t);
    check (ta_eps A = {}) (shows ''no epsilon transitions allowed'');
    check (q \<in> ta_final A) (shows ''final state for '' \<circ> shows t \<circ> shows '' is missing'');
    check (funas_ta A \<subseteq> F) (shows ''the given automaton does not respect the signature'');
    let ts = map (map_r_states Inl) (gi_rules_list F' c t);
    let ms = map (map_r_states Inr) (mp_ta_rules R' F' c);
    check_rules_subseteq ts A';
    check_rules_subseteq ms A';
    let Q = add_rule_states ts (add_rule_states ms []); (*states of internally constructed base TA*)
    let ss = state_substs (r_vars R') Q; (*state substitutions*)
    let D = set (ta_rules_impl' A');
    let D' = inf_step c R' ss D;
    check (D' \<subseteq> D) (shows ''the given tree automaton is not closed under completion rules'');
    check (ta_empty (ta_inter_eps_empty A (ground_instances_ta F c s)))
      (shows ''the given tree automaton does not certify non-reachability'')
  }"

lemma check_etac_nonreachable:
  assumes ok: "isOK (check_etac_nonreachable F a c A R s t)"
  shows "\<not> (\<exists>(\<sigma>::('f::show, 'v::show) subst) \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*)"
proof
  assume "\<exists>(\<sigma>::('f, 'v) subst) \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*"
  then obtain \<sigma> \<tau> :: "('f, 'v) subst" where steps: "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*" by blast
  let ?F = "set F"
  let ?A = "ta_of_ta A"
  let ?T = "ground_instances_ta ?F c t"
  let ?ts = "map (map_r_states Inl) (gi_rules_list F c t)"
  let ?ms = "map (map_r_states Inr) (mp_ta_rules R F c)"
  let ?Q = "add_rule_states ?ts (add_rule_states ?ms [])"
  have aux: "\<And>x A f B. x \<in> A \<Longrightarrow> f ` A \<subseteq> B \<Longrightarrow> f x \<in> B" by blast
  have a: "(a, 0) \<in> ?F" and "(c, 0) \<notin> ?F"
    "\<forall>(l, r) \<in> set R. is_Fun l"
    "\<forall>(l, r) \<in> set R. linear_term l \<and> linear_term r"
    and funas: "funas_trs (set R) \<subseteq> ?F" "funas_term s \<subseteq> ?F" "funas_term t \<subseteq> ?F"
    and "growing (set R)"
    and final: "Inl (star c t) \<in> ta_final ?A"
    and [simp]: "ta_eps ?A = {}"
    and ts: "set ?ts \<subseteq> ta_rules ?A"
    and ms: "set ?ms \<subseteq> ta_rules ?A"
    and inf_step: "inf_step c R (state_substs (r_vars R) ?Q) (ta_rules ?A) \<subseteq> ta_rules ?A"
    and empty: "ta_empty (ta_inter_eps_empty ?A (ground_instances_ta ?F c s))"
    using ok
    by (auto simp: check_etac_nonreachable_def funas_defs linear_trs_def split: tree_automaton.splits)
  moreover then have "funas_ta ?T \<subseteq> ?F" by (simp add: funas_ta_ground_instances_ta)
  ultimately interpret gi_etac c ?F "set R" a ?T
    by (unfold_locales) (auto simp: accessible_ground_instances_ta)
  have subset: "ta_rules cr_ta \<subseteq> ta_rules (ta_of_ta A)"
    using ok
    apply (auto simp: check_etac_nonreachable_def cr_ta_def map_r_states_def split: tree_automaton.splits)
    by (force simp: ground_instances_ta_def map_states_def)+
  have "set ?Q = states cr_ta"
   by (simp add: cr_ta_def states_map_states states_ground_instances_ta [OF a, of c t])
      (auto simp: states_def r_states_def map_r_states_def)
  from overapproximation_of_\<Delta>\<^sub>C [OF subset refl this inf_step]
    have rules: "\<Delta>\<^sub>C \<subseteq> ta_rules ?A" .
  have "ta_lang ?A \<inter> ta_lang (ground_instances_ta ?F c s) = {}"
    using empty [simplified] and intersect_ta and ta_empty by blast
  moreover have "ta_lang C\<^sub>R \<subseteq> ta_lang ?A"
  proof (rule ta_lang_mono)
    have "ta_rules C\<^sub>R \<subseteq> ta_rules ?A" using rules by (auto simp: C\<^sub>R_def)
    moreover have "ta_eps C\<^sub>R \<subseteq> ta_eps ?A" by (auto simp: C\<^sub>R_def cr_ta_def map_states_def)
    moreover have "ta_final C\<^sub>R \<subseteq> ta_final ?A" using final by (auto simp: C\<^sub>R_def cr_ta_def)
    ultimately show "ta_subset C\<^sub>R ?A" by (auto simp: ta_subset_def)
  qed
  moreover have "ground_instances ?F s \<subseteq> ta_lang (ground_instances_ta ?F c s)"
    and "ground_instances ?F t \<subseteq> ta_lang (ground_instances_ta ?F c t)"
    using \<open>funas_term t \<subseteq> ?F\<close> and \<open>funas_term s \<subseteq> ?F\<close>
    by (auto simp: ground_instances_subseteq_lang_ground_instances_ta ta_lang_lang_eq)
  ultimately have *: "ancestors ?F (set R) (ground_instances ?F t) \<inter> ground_instances ?F s = {}"
    apply (simp add: ta_lang_lang_eq)
    apply (subst (asm) etac_ancestors)
    apply (blast dest: ancestors_lang_mono)
    done
  have "\<not> (\<exists>\<sigma> \<tau>. fground ?F (s \<cdot> \<sigma>) \<and> fground ?F (t \<cdot> \<tau>) \<and> (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*)"
    (is "\<not> (\<exists>\<sigma> \<tau>. ?P \<sigma> \<tau>)")
  proof
    assume "\<exists>\<sigma> \<tau>. ?P \<sigma> \<tau>"
    then obtain \<sigma> \<tau> where "?P \<sigma> \<tau>" by blast
    then have "s \<cdot> \<sigma> \<in> ancestors ?F (set R) (ground_instances ?F t)"
      and "s \<cdot> \<sigma> \<in> ground_instances ?F s" by (auto simp: ancestors_def ground_instances_def)
    with * show False by blast
  qed
  with fground_rsteps [OF \<open>(a, 0) \<in> ?F\<close> steps funas] show False by blast
qed

fun map_states_impl
where
  "map_states_impl f (Tree_Automaton qs ts eps) =
    Tree_Automaton (map f qs) (map (map_r_states f) ts) (map (\<lambda>(p, q). (f p, f q)) eps)"

end
