theory Tree_Automata_Wit_Impl
imports
  Tree_Automata_Wit
  Term_Autoref
  "../Framework/Linear_Orders"
  "$AFP/Transitive-Closure/Transitive_Closure_RBT_Impl"
  "$AFP/Collections/Refine_Dflt_Only_ICF"
begin

(*
  Disable the code_printing for some theories that are imported by ICF/Autoref, but are not required
  in our formalization. In particular the code_printing "Array" is problematic, since it is not
  compatible with IsaFoR's code generations setup. (It requires separate Haskell modules for each
  theory).
*)

code_printing code_module "Array" \<rightharpoonup> (Haskell)
code_printing code_module "STArray" \<rightharpoonup> (SML)
code_printing code_module "DiffArray" \<rightharpoonup> (Scala)

code_printing code_module "Uint32" \<rightharpoonup> (SML)
code_printing code_module "Uint32" \<rightharpoonup> (Haskell)
code_printing code_module "Uint32" \<rightharpoonup> (OCaml)
code_printing code_module "Uint32" \<rightharpoonup> (Scala)

code_printing code_module "Bits_Integer" \<rightharpoonup> (SML)
code_printing code_module "Data_Bits" \<rightharpoonup> (Haskell)
code_printing code_module "Bits_Integer" \<rightharpoonup> (OCaml)
code_printing code_module "Bits_Integer" \<rightharpoonup> (Scala)

(* The notation for TA_rule clashes with the one for fun_rel. *)
no_notation TA_rule ("_ _ \<rightarrow> _" [51, 51, 51] 52)

subsection {* Auxiliary Autoref setup for some basic functions *}

lemma option_bind_ref[autoref_rules, param]:
  "(Option.bind, Option.bind) \<in>  \<langle>Ra\<rangle>option_rel \<rightarrow> (Ra \<rightarrow> \<langle>Rb\<rangle>option_rel) \<rightarrow> \<langle>Rb\<rangle>option_rel"
unfolding Option.bind_def by parametricity

lemma option_mapm_ref[autoref_rules, param]:
  "(mapM, mapM) \<in> (Ra \<rightarrow> \<langle>Rb\<rangle>option_rel) \<rightarrow> \<langle>Ra\<rangle>list_rel \<rightarrow> \<langle>\<langle>Rb\<rangle>list_rel\<rangle>option_rel"
proof (intro fun_relI, goal_cases)
  case (1 f f' l l')
  thus ?case by (induction l' arbitrary: l, auto elim!: list_relE option_relE) parametricity
qed

context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "\<forall>x\<in>set l. P x \<equiv> list_all_rec$P$l"
by (simp add: list_all_rec_eq)
end

lemma list_all_rec_autoref[autoref_rules]:
  "(list_all_rec, list_all_rec) \<in> (A \<rightarrow> bool_rel) \<rightarrow> \<langle>A\<rangle>list_rel \<rightarrow> bool_rel"
unfolding list_all_rec_def
by parametricity

fun list_all2_rec where
  "list_all2_rec f (x#xs) (y#ys) = ((f x y) \<and> list_all2_rec f xs ys)"
| "list_all2_rec f [] ys = (ys = [])"
| "list_all2_rec f xs [] = (xs = [])"

lemma list_all2_autoref[autoref_rules]:
  "(list_all2, list_all2) \<in> (J \<rightarrow> K \<rightarrow> bool_rel) \<rightarrow> \<langle>J\<rangle>list_rel \<rightarrow> \<langle>K\<rangle>list_rel \<rightarrow> bool_rel"
proof (intro fun_relI, goal_cases)
case (1 f f' l1 l1' l2 l2')
  thus ?case proof (induction f l1 l2 arbitrary: l1' l2' rule: list_all2_rec.induct)
  case prems: (1 f x xs y ys)
    from this obtain x' xs' y' ys' where
      xs: "l1' = x'#xs'" "(xs,xs') \<in> \<langle>J\<rangle>list_rel" and x: "(x,x') \<in> J" and 
      ys: "l2' = y'#ys'" "(ys,ys') \<in> \<langle>K\<rangle>list_rel" and y: "(y,y') \<in> K" by (auto elim!: list_relE3)
    with prems have "(list_all2 f xs ys, list_all2 f' xs' ys') \<in> bool_rel" by blast
    with prems x y show ?case by (simp only: list.simps xs ys, parametricity)
  qed auto
qed

lemma [autoref_itype]:
  "list_all_rec ::\<^sub>i (I \<rightarrow>\<^sub>i i_bool) \<rightarrow>\<^sub>i \<langle>I\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i i_bool"
  "list_all2 ::\<^sub>i (J \<rightarrow>\<^sub>i K \<rightarrow>\<^sub>i i_bool) \<rightarrow>\<^sub>i \<langle>J\<rangle>\<^sub>ii_list  \<rightarrow>\<^sub>i \<langle>K\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i i_bool"
by simp_all

(*
  The "map iteration theorem" from the autoref framework is not general enough for our use case,
  since it does not allow the value type to be refined. (i.e. the refinement relation must be Id). 
  This is problematic, since we use nested maps to represent relations, were the outer as well as
  the inner one is refined. Therefore, we prove the following, stronger one, which allows
  arbitrary refinement relations for values (Rv).
*)

thm poly_map_iteratei.autoref_iteratei
lemma (in poly_map_iteratei) autoref_iteratei[autoref_ga_rules]: 
  assumes ID: "REL_IS_ID Rk"
  shows "is_map_to_list Rk Rv rel (it_to_list iteratei)"
unfolding is_map_to_sorted_list_def is_map_to_list_def it_to_sorted_list_def rel_def
proof (intro allI impI conjI, elim relcompE, insert assms, simp)
  fix m m'
  assume "(m, m') \<in> br \<alpha> invar"
  hence I: "invar m" and M': "m' = \<alpha> m" unfolding br_def by (simp_all add: br_def)

  have [simp]: "\<And>c. (\<lambda>(_,_). c) = (\<lambda>_. c)" by auto

  from map_it_to_list_genord_correct[where it = iteratei, 
    where R="\<lambda>_ _. True", simplified, OF 
    iteratei_correct[OF I, unfolded set_iterator_def]
  ] have 
      M: "Map.map_of (it_to_list iteratei m) = \<alpha> m"
      and D: "distinct (List.map fst (it_to_list iteratei m))"
    by (simp_all)

  from M have mts: "map_to_set m' = set (it_to_list iteratei m)"
    by (simp add: M' map_of_map_to_set[OF D])

  fix m''
  assume "(m', m'') \<in> Id \<rightarrow> \<langle>Rv\<rangle>option_rel"
  hence opt: "\<And>k. (m' k, m'' k) \<in> \<langle>Rv\<rangle>option_rel" by (auto dest: fun_relD)
  hence dom: "dom m' = dom m''" by (force simp: dom_def elim: option_relE)
  {
    fix k
    assume "k \<in> dom m'"
    from this obtain v where v: "m' k = Some v" by (auto simp: dom_def)
    from this obtain v' where v': "m'' k = Some v'" "(v,v') \<in> Rv" using opt[of k] by (auto elim: option_relE)
    from v v' have "\<exists>v. m' k = Some (fst v) \<and> m'' k = Some (snd v) \<and> v \<in> Rv" by auto 
  }
  hence "\<forall>k. \<exists>v. k \<in> dom m' \<longrightarrow> m' k = Some (fst v) \<and> m'' k = Some (snd v) \<and> v \<in> Rv" by blast
  with choice[OF this] obtain f where
    f: "\<And>k. k \<in> dom m' \<Longrightarrow> m' k = Some (fst (f k)) \<and> m'' k = Some (snd (f k)) \<and> (f k) \<in> Rv"
  by blast
  let ?f = "\<lambda>(k,v). (k, snd (f k))"
  have mts': "map_to_set m'' = set (map ?f (it_to_list iteratei m))"
  unfolding set_map mts[symmetric] proof
    show "map_to_set m'' \<subseteq> ?f ` map_to_set m'"
    proof (rule subrelI)
      fix x y assume "(x, y) \<in> map_to_set m''"
      hence the: "m'' x = Some y" and "x \<in> dom m''" by (auto simp: map_to_set_def)
      hence "x \<in> dom m'" unfolding dom by blast
      from f[OF this] the have "snd (f x) = y" "m' x = Some (fst (f x))" by auto
      thus "(x, y) \<in> ?f ` map_to_set m'" by (intro image_eqI[where x = "(x, fst (f x))"], auto)
    qed
    show "?f ` map_to_set m' \<subseteq> map_to_set m''"
    proof (rule subrelI)
      fix x y assume *: "(x,y) \<in> ?f ` map_to_set m'"
      hence "x \<in> dom m'" by (auto simp: map_to_set_def)
      from f[OF this] * have " m'' x = Some y" by auto
      thus "(x,y) \<in> map_to_set m''" by (auto simp: map_to_set_def)
    qed
  qed
  have fst: "fst o ?f = fst" by (auto simp: fun_eq_iff)
  from D have dist: "distinct (map fst (map ?f (it_to_list iteratei m)))" by (auto simp: fst)
  have "(it_to_list iteratei m, map ?f (it_to_list iteratei m)) \<in> \<langle>Id \<times>\<^sub>r Rv\<rangle>list_rel"
  unfolding list_rel_def proof (auto simp add: list_all2_conv_all_nth split: prod.split)
    fix i k v
    assume "i < length (it_to_list iteratei m)" "it_to_list iteratei m ! i = (k,v)"  
    with mts nth_mem[OF this(1)] have "(k,v) \<in> map_to_set m'" by auto
    hence "k \<in> dom m'" "m' k = Some v" by (auto simp: map_to_set_def)
    with f[OF this(1)] show "(v, snd (f k)) \<in> Rv" by auto
  qed
  from this mts' distinct_mapI[OF dist] show
    "\<exists>l'. (it_to_list iteratei m, l') \<in> \<langle>Id \<times>\<^sub>r Rv\<rangle>list_rel \<and> distinct l' \<and> map_to_set m'' = set l'"
  by (intro exI[of _ "map ?f (it_to_list iteratei m)"], blast+)
qed

subsection {* Autoref setup for @{typ "('q,'f) ta_rule"} *}

definition ta_rule_rel :: "_ \<Rightarrow> _ \<Rightarrow> ((_,_) ta_rule \<times> (_,_) ta_rule) set" where
  ta_rule_rel_def_internal: "ta_rule_rel R_sym R_states \<equiv>
  { (TA_rule f qs q, TA_rule f' qs' q') | f f' qs qs' q q'.
    (f, f') \<in> R_sym \<and>
    (qs, qs') \<in> \<langle>R_states\<rangle>list_rel \<and>
    (q, q') \<in> R_states}"

lemma ta_rule_rel_def[refine_rel_defs]: 
  "\<langle>R_sym, R_states\<rangle>ta_rule_rel \<equiv>
  { (TA_rule f qs q, TA_rule f' qs' q') | f f' qs qs' q q'.
    (f, f') \<in> R_sym \<and>
    (qs, qs') \<in> \<langle>R_states\<rangle>list_rel \<and>
    (q, q') \<in> R_states }"    
  by (simp add: ta_rule_rel_def_internal relAPP_def)

lemma ta_rule_rel_sv[relator_props]:
  assumes "single_valued R_sym" and states: "single_valued R_states"
  shows "single_valued (\<langle>R_sym, R_states\<rangle>ta_rule_rel)"
using assms states[THEN list_rel_sv]
by (auto simp add: ta_rule_rel_def intro: single_valuedI dest: single_valuedD)

lemma ta_rule_rel_id[relator_props]:
  assumes "R_sym = Id" and "R_states = Id"
    shows "\<langle>R_sym, R_states\<rangle>ta_rule_rel = Id"
using assms by (auto simp add: ta_rule_rel_def intro: ta_rule.exhaust)

lemma ta_rule_rel_id_simp[simp]:
  shows "\<langle>Id, Id\<rangle>ta_rule_rel = Id" by tagged_solver

lemma ta_rule_rel_mono[relator_props]:
  assumes "R_sym \<subseteq> R_sym'" and states: "R_states \<subseteq> R_states'"
  shows "\<langle>R_sym, R_states\<rangle>ta_rule_rel \<subseteq> \<langle>R_sym', R_states'\<rangle>ta_rule_rel"
using assms states[THEN list_rel_mono] by (auto simp add: ta_rule_rel_def)

lemma ta_rule_rel_range[relator_props]:
  assumes "Range R_sym = UNIV" and states: "Range R_states = UNIV"
    shows "Range (\<langle>R_sym, R_states\<rangle>ta_rule_rel) = UNIV"
proof (rule set_eqI, clarsimp)
  fix x
  show "x \<in> Range (\<langle>R_sym, R_states\<rangle>ta_rule_rel)"
  using assms states[THEN list_rel_range] by (cases x, auto simp add: ta_rule_rel_def Range_iff)
qed

lemma ta_rule_autoref[autoref_rules, param]:
  "(TA_rule, TA_rule) \<in> F \<rightarrow> \<langle>Q\<rangle>list_rel \<rightarrow> Q \<rightarrow> \<langle>F, Q\<rangle>ta_rule_rel"
  "(case_ta_rule, case_ta_rule) \<in> (F \<rightarrow> \<langle>Q\<rangle>list_rel \<rightarrow> Q \<rightarrow> X) \<rightarrow> \<langle>F, Q\<rangle>ta_rule_rel \<rightarrow> X"
  "(r_rhs, r_rhs) \<in> \<langle>F, Q\<rangle>ta_rule_rel \<rightarrow> Q"
  "(r_lhs_states, r_lhs_states) \<in> \<langle>F, Q\<rangle>ta_rule_rel \<rightarrow> \<langle>Q\<rangle>list_rel"
by (force simp: ta_rule_rel_def dest: fun_relD)+

consts i_ta_rule :: "interface => interface => interface"
lemmas [autoref_rel_intf] = REL_INTFI[of ta_rule_rel i_ta_rule]

lemma [autoref_itype]:
  "TA_rule ::\<^sub>i F \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i Q \<rightarrow>\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_ta_rule"
  "case_ta_rule ::\<^sub>i (F \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i Q \<rightarrow>\<^sub>i X) \<rightarrow>\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_ta_rule \<rightarrow>\<^sub>i X"
  "r_rhs ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_ta_rule \<rightarrow>\<^sub>i Q"
  "r_lhs_states ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_ta_rule \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_list"
by simp_all

lemma transfer_term[refine_transfer]:
  assumes "\<And>x. \<alpha> (fa x) \<le> Fa x"
  assumes "\<And>x y. \<alpha> (fb x y) \<le> Fb x y"
  shows "\<alpha> (case_term fa fb x) \<le> case_term Fa Fb x"
using assms by (auto split: term.split)

lemma transfer_ta_rule[refine_transfer]:
  assumes "\<And>x y z. \<alpha> (fa x y z) \<le> Fa x y z"
  shows "\<alpha> (case_ta_rule fa x) \<le> case_ta_rule Fa x"
using assms by (auto split: ta_rule.split)

subsection {* Autoref setup for @{typ "('q,'f) ta"} *}

definition ta_idx :: "(_,_)ta \<Rightarrow> _"  where
  "ta_idx TA \<equiv> \<lambda>(f,n).
    let rs = {(qs,q) | f' qs q. TA_rule f qs q \<in> ta_rules TA \<and> f' = f \<and> length qs = n} in 
      if rs = {} then None else Some rs"

abbreviation "ta_idx_rel F Q Rm Rs \<equiv> \<langle>F \<times>\<^sub>r nat_rel, \<langle>\<langle>Q\<rangle>list_rel \<times>\<^sub>r Q\<rangle>Rs\<rangle>Rm"

definition TA_rel where
  TA_rel_internal_def: "TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl R_sym R_states \<equiv> 
  {((F,R,E,rhss,idx,det,efcl,eicl),TA).
    (F, ta_final TA) \<in> \<langle>R_states\<rangle>R_final \<and> 
    (R, ta_rules TA) \<in> \<langle>\<langle>R_sym,R_states\<rangle>ta_rule_rel\<rangle>R_rules \<and>
    (E, ta_eps TA) \<in> \<langle>R_states \<times>\<^sub>r R_states\<rangle>R_eps \<and>
    (rhss, ta_rhs_states TA) \<in> \<langle>R_states\<rangle>R_rhss \<and>
    (idx, ta_idx TA) \<in> ta_idx_rel R_sym R_states R_idx R_idxs \<and>
    (det, ta_det TA) \<in> Id \<and>
    (efcl, \<lambda>q. {p. (q, p) \<in> (ta_eps TA)\<^sup>*}) \<in> \<langle>R_states,\<langle>R_states\<rangle>R_cl\<rangle>fun_rel \<and>
    (eicl, \<lambda>q. {p. (p, q) \<in> (ta_eps TA)\<^sup>*}) \<in> \<langle>R_states,\<langle>R_states\<rangle>R_cl\<rangle>fun_rel}"

lemma TA_rel_def[refine_rel_defs]:
  "\<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<equiv>
  {((F,R,E,rhss,idx,det,efcl,eicl),TA).
    (F, ta_final TA) \<in> \<langle>R_states\<rangle>R_final \<and> 
    (R, ta_rules TA) \<in> \<langle>\<langle>R_sym,R_states\<rangle>ta_rule_rel\<rangle>R_rules \<and>
    (E, ta_eps TA) \<in> \<langle>R_states \<times>\<^sub>r R_states\<rangle>R_eps \<and>
    (rhss, ta_rhs_states TA) \<in> \<langle>R_states\<rangle>R_rhss \<and>
    (idx, ta_idx TA) \<in> ta_idx_rel R_sym R_states R_idx R_idxs \<and>
    (det, ta_det TA) \<in> Id \<and>
    (efcl, \<lambda>q. {p. (q, p) \<in> (ta_eps TA)\<^sup>*}) \<in> \<langle>R_states,\<langle>R_states\<rangle>R_cl\<rangle>fun_rel \<and>
    (eicl, \<lambda>q. {p. (p, q) \<in> (ta_eps TA)\<^sup>*}) \<in> \<langle>R_states,\<langle>R_states\<rangle>R_cl\<rangle>fun_rel}"
  by (simp add: TA_rel_internal_def relAPP_def)

lemma NFA_rel_sv[relator_props]:
  assumes "single_valued (\<langle>R_states\<rangle>R_final)"
  assumes "single_valued (\<langle>\<langle>R_sym,R_states\<rangle>ta_rule_rel\<rangle>R_rules)"
  assumes "single_valued (\<langle>R_states \<times>\<^sub>r R_states\<rangle>R_eps)"
  shows "single_valued (\<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl)"
unfolding TA_rel_def
proof (intro single_valuedI allI impI, clarify, goal_cases)
  case (1 _ _ _ _ _ _ _ _ y z )
    thus "y = z" by (cases y, cases z, auto dest: assms[THEN single_valuedD])
qed

fun ta_final_impl where "ta_final_impl (F,R,E,rhs,idx,det,efcl,eicl) = F"
fun ta_rules_impl where "ta_rules_impl (F,R,E,rhs,idx,det,efcl,eicl) = R"
fun ta_eps_impl where "ta_eps_impl (F,R,E,rhs,idx,det,efcl,eicl) = E"
fun ta_eps_cl_impl where "ta_eps_cl_impl (F,R,E,rhs,idx,det,efcl,eicl) q = efcl q"
fun ta_eps_icl_impl where "ta_eps_icl_impl (F,R,E,rhs,idx,det,efcl,eicl) q = eicl q"
fun ta_rhs_states_impl where "ta_rhs_states_impl (F,R,E,rhs,idx,det,efcl,eicl) = rhs"
fun ta_idx_impl where "ta_idx_impl (F,R,E,rhs,idx,det,efcl,eicl) = idx"
fun ta_det_impl where "ta_det_impl (F,R,E,rhs,idx,det,x) = det"

lemma ta_impl_autoref[autoref_rules]:
  "(ta_final_impl, ta_final) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> \<langle>R_states\<rangle>R_final"
  "(ta_rules_impl, ta_rules) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> \<langle>\<langle>R_sym,R_states\<rangle>ta_rule_rel\<rangle>R_rules"
  "(ta_eps_impl, ta_eps) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> \<langle>R_states \<times>\<^sub>r R_states\<rangle>R_eps"
  "(ta_rhs_states_impl, ta_rhs_states) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> \<langle>R_states\<rangle>R_rhss"
  "(ta_idx_impl, ta_idx) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> ta_idx_rel R_sym R_states R_idx R_idxs"
  "(ta_det_impl, ta_det) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> Id"
by (auto simp: TA_rel_def)

definition [simp]: "op_ta_eps_cl (TA::(_,_)ta) q \<equiv> {p. (q, p) \<in> (ta_eps TA)\<^sup>*}"
definition [simp]: "op_ta_eps_icl (TA::(_,_)ta) q \<equiv> {p. (p, q) \<in> (ta_eps TA)\<^sup>*}"

context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "{p. (q, p) \<in> (ta_eps TA)\<^sup>*} \<equiv> op_ta_eps_cl$TA$q"
  "{p. (p, q) \<in> (ta_eps TA)\<^sup>*} \<equiv> op_ta_eps_icl$TA$q"
by simp_all
end

lemma eps_cl_refine[autoref_rules]:
  "(ta_eps_cl_impl, op_ta_eps_cl) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> R_states \<rightarrow> \<langle>R_states\<rangle>R_cl"
  "(ta_eps_icl_impl, op_ta_eps_icl) \<in> \<langle>R_sym, R_states\<rangle>TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl \<rightarrow> R_states \<rightarrow> \<langle>R_states\<rangle>R_cl"
using assms by (auto simp: TA_rel_def dest: fun_relD)

consts i_TA :: "interface => interface => interface"
lemmas [autoref_rel_intf] =
    REL_INTFI[of "TA_rel R_final R_rules R_eps R_rhss R_idx R_idxs R_cl" i_TA]
      for R_final R_rules R_eps R_rhss R_idx R_idxs R_cl

lemma [autoref_itype]:
  "op_ta_eps_cl ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i Q \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_set"
  "op_ta_eps_icl ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i Q \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_set"
  "ta_rules ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i \<langle>\<langle>F,Q\<rangle>\<^sub>ii_ta_rule\<rangle>\<^sub>ii_set"
  "ta_eps ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i \<langle>\<langle>Q,Q\<rangle>\<^sub>ii_prod\<rangle>\<^sub>ii_set"
  "ta_final ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_set"
  "ta_rhs_states ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i \<langle>Q\<rangle>\<^sub>ii_set"
  "ta_idx ::\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA \<rightarrow>\<^sub>i \<langle>\<langle>Q,i_nat\<rangle>\<^sub>ii_prod, \<langle>\<langle>\<langle>Q\<rangle>\<^sub>ii_list,Q\<rangle>\<^sub>ii_prod\<rangle>\<^sub>ii_set\<rangle>\<^sub>ii_map"
by simp_all

derive compare_order ta_rule
abbreviation "dflt_TA_rel \<equiv> TA_rel rs.rel rs.rel ls.rel rs.rel rm.rel rs.rel rs.rel"

lemma dflt_NFA_rel_sv[relator_props]:
  assumes [relator_props]:
    "single_valued R_states"
    "single_valued R_sym"
  shows "single_valued (\<langle>R_sym, R_states\<rangle>dflt_TA_rel)"
by tagged_solver

subsection {* Refinement of auxiliary functions *}

definition [simp]: "op_union_image S f \<equiv> \<Union>(f ` S)"
definition "union_image_ref S f \<equiv> FOREACH S (\<lambda>s r. RETURN (f s \<union> r)) {}"

context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "\<Union>(f ` S) \<equiv> op_union_image$S$f"
by simp_all
end

lemma union_image_ref:
  assumes "finite S"
    shows "union_image_ref S f \<le> RETURN(op_union_image S f)"
unfolding union_image_ref_def
by (refine_vcg FOREACH_rule[where I = "\<lambda>it r. r = op_union_image (S - it) f"], auto intro: assms)

schematic_goal union_image_rs_aux:
  assumes [autoref_rules]: "(S',S) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(f',f) \<in> Id \<rightarrow> \<langle>Id\<rangle>rs.rel"
  shows "(?f::?'a, union_image_ref S f) \<in> ?R"
unfolding union_image_ref_def[abs_def]
by (autoref_monadic (plain))

concrete_definition union_image_rs_code uses union_image_rs_aux
thm union_image_rs_code_def

lemma union_image_rs_autoref[autoref_rules]:
  assumes "PREFER_id A" "PREFER_id B"
    shows "(union_image_rs_code,op_union_image) \<in> \<langle>A\<rangle>rs.rel \<rightarrow> (A \<rightarrow> \<langle>B\<rangle>rs.rel) \<rightarrow> \<langle>B\<rangle>rs.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "A = Id" "B = Id" by simp_all
  case rel: (1 S S')
    hence fin: "finite S'" by (simp add: br_def)
    note union_image_rs_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note union_image_ref[OF fin]
    finally show ?case by simp
qed

definition [simp]: "op_image S f \<equiv> f ` S"
definition "image_ref S f \<equiv> FOREACH S (\<lambda>s r. RETURN (insert (f s) r)) {}"

context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "f ` S \<equiv> op_image$S$f"
by simp_all
end

lemma image_ref:
  assumes "finite S"
    shows "image_ref S f \<le> RETURN(op_image S f)"
unfolding image_ref_def
by (refine_vcg FOREACH_rule[where I = "\<lambda>it r. r = op_image (S - it) f"], auto intro: assms)

schematic_goal image_ls_aux:
  assumes [autoref_rules]: "(S',S) \<in> \<langle>Id\<rangle>ls.rel"
  assumes [autoref_rules]: "(f',f) \<in> Id \<rightarrow> Id"
  shows "(?f, image_ref S f) \<in> \<langle>\<langle>Id\<rangle>ls.rel\<rangle>nres_rel"
unfolding image_ref_def[abs_def]
by (autoref_monadic (plain))

concrete_definition image_ls_code uses image_ls_aux

lemma image_ls_autoref[autoref_rules]:
  assumes "PREFER_id A" "PREFER_id B"
    shows "(image_ls_code,op_image) \<in> \<langle>A\<rangle>ls.rel \<rightarrow> (A \<rightarrow> B) \<rightarrow> \<langle>B\<rangle>ls.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "A = Id" "B = Id" by simp_all
  case rel: (1 S S')
    hence fin: "finite S'" by (simp add: br_def)
    note image_ls_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note image_ref[OF fin]
    finally show ?case by simp
qed

schematic_goal image_rs_aux:
  assumes [autoref_rules]: "(S',S) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(f',f) \<in> Id \<rightarrow> Id"
  shows "(?f, image_ref S f) \<in> \<langle>\<langle>Id\<rangle>rs.rel\<rangle>nres_rel"
unfolding image_ref_def[abs_def]
by (autoref_monadic (plain))

concrete_definition image_rs_code uses image_rs_aux

lemma image_rs_autoref[autoref_rules]:
  assumes "PREFER_id A" "PREFER_id B"
    shows "(image_rs_code,op_image) \<in> \<langle>A\<rangle>rs.rel \<rightarrow> (A \<rightarrow> B) \<rightarrow> \<langle>B\<rangle>rs.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "A = Id" "B = Id" by simp_all
  case rel: (1 S S')
    hence fin: "finite S'" by (simp add: br_def)
    note image_rs_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note image_ref[OF fin]
    finally show ?case by simp
qed

lemma swap_converse: "prod.swap ` A = A\<inverse>" by auto

definition ta_idx_rhs_init where
  "ta_idx_rhs_init rs det efcl \<equiv>
    FOREACH rs (\<lambda>r (idx,det,rhs). RETURN (case r of TA_rule f qs q \<Rightarrow>
      let n = length qs in
        (case idx (f,n) of
          None \<Rightarrow> (idx((f,n) \<mapsto> {(qs,q)}), det, efcl q \<union> rhs)
        | Some rs \<Rightarrow>
            (idx((f,n) \<mapsto> insert (qs,q) rs),
            if det then \<forall>(qs',q') \<in> rs. qs \<noteq> qs' else False,
            efcl q \<union> rhs)))
    ) (Map.empty, det, {})"

lemma r_rhs_unfold: "r_rhs = case_ta_rule (\<lambda>_ _ q. q)" by (auto simp: fun_eq_iff split: ta_rule.split)

lemma ta_diff_aux_simps[simp]:
  "ta_idx (ta_diff TA (ta_rules TA)) = Map.empty"
  "ta_det (ta_diff TA (ta_rules TA)) = (ta_eps TA = {})"
unfolding ta_diff_def ta_idx_def ta_det_def by auto

lemma ta_det_it:
  assumes "TA_rule f qs q \<in> it" "it \<subseteq> ta_rules TA"
    shows "ta_det (ta_diff TA (it - {TA_rule f qs q})) =
            (ta_det (ta_diff TA it) \<and>
              (case (ta_idx (ta_diff TA it)) (f, length qs) of
                None \<Rightarrow> True 
              | Some rs \<Rightarrow> \<forall>(qs',q') \<in> rs. qs \<noteq> qs'))"
using assms by (simp add: it_step_insert_iff ta_det_def ta_idx_def ta_diff_rules, auto) blast

lemma ta_idx_it_none:
  assumes "TA_rule f qs q \<in> it" "it \<subseteq> ta_rules TA"
      and "ta_idx (ta_diff TA it) (f, length qs) = None"
    shows "ta_idx (ta_diff TA (it - {TA_rule f qs q})) =
            ta_idx (ta_diff TA it)((f, length qs) \<mapsto> {(qs,q)})"
using assms by (auto simp: it_step_insert_iff ta_idx_def ta_diff_rules fun_eq_iff split: if_splits)

lemma ta_idx_it_some:
  assumes "TA_rule f qs q \<in> it" "it \<subseteq> ta_rules TA"
      and "ta_idx (ta_diff TA it) (f, length qs) = Some rs"
    shows "ta_idx (ta_diff TA (it - {TA_rule f qs q})) =
            ta_idx (ta_diff TA it)((f, length qs) \<mapsto> insert (qs,q) rs)"
using assms by (auto simp: it_step_insert_iff ta_idx_def ta_diff_rules fun_eq_iff split: if_splits)

lemma ta_rhs_states_it:
  assumes "TA_rule f qs q \<in> it" "it \<subseteq> ta_rules TA"
    shows "ta_rhs_states (ta_diff TA (it - {TA_rule f qs q})) =
            ta_rhs_states (ta_diff TA it) \<union> eps_cl TA q"
using assms by (auto simp: it_step_insert_iff ta_rhs_states_def ta_diff_rules)

lemma ta_idx_rhs_init:
  assumes fin: "finite (ta_rules TA)"
      and eps: "\<And>q. efcl q \<equiv> eps_cl TA q"
    shows "ta_idx_rhs_init (ta_rules TA) (ta_eps TA = {}) efcl \<le> RETURN(ta_idx TA, ta_det TA, ta_rhs_states TA)"
proof -
  note rules =
    FOREACH_rule[where I = 
        "\<lambda>it (idx,det,rhs).
          let TA = (ta_diff TA it) in
            idx = ta_idx TA \<and> det = ta_det TA \<and> rhs = ta_rhs_states TA", OF fin, unfolded Let_def]
    ta_rule_rule
  note simps = Let_def ta_det_it ta_idx_it_none ta_idx_it_some ta_rhs_states_it
  note splits = ta_rule.splits option.splits if_splits
  show ?thesis unfolding ta_idx_rhs_init_def eps
  by (refine_vcg rules; auto simp: simps split: splits)
qed

abbreviation "memo_rtrancl_rel Q \<equiv> Q \<rightarrow> \<langle>Q\<rangle>rs.rel"

schematic_goal ta_idx_rhs_init_aux:
  assumes "PREFER_id F" "PREFER_id Q"
  shows "(?f, ta_idx_rhs_init) \<in>
          \<langle>\<langle>F,Q\<rangle>ta_rule_rel\<rangle>rs.rel \<rightarrow> Id \<rightarrow> memo_rtrancl_rel Q \<rightarrow> \<langle>ta_idx_rel F Q rm.rel rs.rel \<times>\<^sub>r Id \<times>\<^sub>r \<langle>Q\<rangle>rs.rel \<rangle>nres_rel"
unfolding ta_idx_rhs_init_def[abs_def] assms[THEN PREFER_id_D]
by autoref

concrete_definition ta_idx_rhs_init_impl uses ta_idx_rhs_init_aux
lemmas [autoref_rules] = ta_idx_rhs_init_impl.refine

schematic_goal ta_idx_rhs_init_transfer_aux: 
  "RETURN ?c \<le> ta_idx_rhs_init_impl rs det efcl"
unfolding ta_idx_rhs_init_impl_def
by (refine_transfer (post))

concrete_definition ta_idx_rhs_init_code for rs efcl uses ta_idx_rhs_init_transfer_aux
lemmas [refine_transfer] = ta_idx_rhs_init_code.refine
thm ta_idx_rhs_init_code_def

definition memo_rtrancl where
  "memo_rtrancl S x \<equiv> {y. (x, y) \<in> S\<^sup>*}"

definition memo_rbt_rtrancl_opt where
  "memo_rbt_rtrancl_opt E \<equiv>
    if ls.isEmpty E then (\<lambda>q. rs.sng q) else memo_rbt_rtrancl (ls.to_list E)"

lemma memo_rtrancl_autoref[autoref_rules]:
  assumes "PREFER_id R"
    shows "(memo_rbt_rtrancl_opt, memo_rtrancl) \<in> \<langle>R \<times>\<^sub>r R\<rangle>ls.rel \<rightarrow> R \<rightarrow> \<langle>R\<rangle>rs.rel" 
using assms
by (auto simp: memo_rtrancl_def memo_rbt_rtrancl_opt_def br_def ls.correct rs.correct memo_rbt_rtrancl split: if_splits)

definition ta_make where
  "ta_make F R E \<equiv>
    let efcl = memo_rtrancl E in do {
    (idx,det,rhs) \<leftarrow> ta_idx_rhs_init R (E = {}) efcl;
      RETURN
      (F, R, E,
        rhs, idx, det,
        efcl, memo_rtrancl (prod.swap ` E))
    }"

lemma ta_make_correct:
  assumes "finite R"
  shows "ta_make F R E \<le>
          SPEC(\<lambda>(F',R',E',rhs,idx,det,efcl,eicl). let TA = ta.make F R E in
            F = F' \<and> R = R' \<and> E = E' \<and>
            rhs = ta_rhs_states TA \<and> idx = ta_idx TA \<and> det = ta_det TA \<and>
            efcl = eps_cl TA \<and> eicl = eps_icl TA)"
proof -
  note ta.make_def[simp]
  from assms have fin: "finite (ta_rules (ta.make F R E))" by simp
  have eps: "\<And>q. memo_rtrancl E q \<equiv> eps_cl (ta.make F R E) q" by (simp add: memo_rtrancl_def)
  note rule = ta_idx_rhs_init[OF fin eps, simplified]
  show ?thesis unfolding ta_make_def
  by (refine_vcg rule[THEN order_trans], auto simp: memo_rtrancl_def[abs_def] swap_converse)
     (auto simp: fun_eq_iff intro: rtrancl_converseI dest: rtrancl_converseD)
qed

schematic_goal ta_make_aux:
  assumes [autoref_rules]: "(F',F) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(R',R) \<in> \<langle>\<langle>Id,Id\<rangle>ta_rule_rel\<rangle>rs.rel"
  assumes [autoref_rules]: "(E',E) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>ls.rel"
  shows "(?f, ta_make F R E) \<in> 
    \<langle>\<langle>Id\<rangle>rs.rel \<times>\<^sub>r \<langle>\<langle>Id,Id\<rangle>ta_rule_rel\<rangle>rs.rel \<times>\<^sub>r \<langle>Id \<times>\<^sub>r Id\<rangle>ls.rel \<times>\<^sub>r
     \<langle>Id\<rangle>rs.rel \<times>\<^sub>r ta_idx_rel Id Id rm.rel rs.rel \<times>\<^sub>r Id  \<times>\<^sub>r
     memo_rtrancl_rel Id \<times>\<^sub>r memo_rtrancl_rel Id\<rangle>nres_rel"
unfolding ta_make_def[abs_def] prod.swap_def[abs_def]
by (autoref_monadic (plain))

concrete_definition ta_make_code uses ta_make_aux

lemma ta_make_autoref[autoref_rules]:
  assumes "PREFER_id Q" "PREFER_id F"
  shows "(ta_make_code, ta.make) \<in> \<langle>Q\<rangle>rs.rel \<rightarrow> \<langle>\<langle>F,Q\<rangle>ta_rule_rel\<rangle>rs.rel \<rightarrow> \<langle>Q \<times>\<^sub>r Q\<rangle>ls.rel \<rightarrow> \<langle>F,Q\<rangle>dflt_TA_rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "Q = Id" "F = Id" by simp_all
  case rel: (1 _ _ R R')
    hence fin: "finite R'" by (auto simp add: br_def)
    note ta_make_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note ta_make_correct[OF fin]
    finally show ?case
      by (auto simp add: TA_rel_def RETURN_RES_refine_iff Let_def ta.make_def elim!: prod_relE)
         (auto dest!: fun_relD)
qed

lemma [autoref_itype]:
  "ta.make ::\<^sub>i \<langle>Q\<rangle>\<^sub>ii_set \<rightarrow>\<^sub>i \<langle>\<langle>F,Q\<rangle>\<^sub>ii_ta_rule\<rangle>\<^sub>ii_set \<rightarrow>\<^sub>i \<langle>\<langle>Q,Q\<rangle>\<^sub>ii_prod\<rangle>\<^sub>ii_set \<rightarrow>\<^sub>i \<langle>F,Q\<rangle>\<^sub>ii_TA"
by simp

definition has_res_wit where
  "has_res_wit m r \<equiv> case r of TA_rule f qs q \<Rightarrow>
    do { ts \<leftarrow> mapM (\<lambda>x. op_map_lookup x m) qs; Some (q, Fun f ts) }"

definition next_res_wit_refine where
  "next_res_wit_refine rs m \<equiv> FOREACH\<^sub>C rs (\<lambda>w. w = None) (\<lambda>r _. RETURN (has_res_wit m r)) None"

lemma next_res_wit_refine:
  assumes "finite rs"
    shows "next_res_wit_refine rs m \<le> SPEC(next_res_wit_inv rs m)"
unfolding next_res_wit_refine_def next_res_wit_inv_def
proof (refine_vcg FOREACHc_rule[where I="\<lambda>it. next_res_wit_inv (rs - it) m"], goal_cases)
  case prems: (3 r it w)
    obtain f qs q where r: "r = TA_rule f qs q" by (cases r)
    from prems show ?case unfolding r next_res_wit_inv_def
      by (cases "mapM m qs", simp_all add: has_res_wit_def, auto simp: mapM_None dest: mapM_Some)
qed (auto simp: next_res_wit_inv_def assms)

schematic_goal next_res_wit_refine_aux:
  assumes [autoref_rules]: "(R',R) \<in> \<langle>\<langle>Id,Id\<rangle>ta_rule_rel\<rangle>rs.rel"
  assumes [autoref_rules]: "(M',M) \<in> \<langle>Id,\<langle>Id,V\<rangle>term_rel\<rangle>rm.rel"
  shows "(?f, next_res_wit_refine R M) \<in> \<langle>\<langle>Id \<times>\<^sub>r \<langle>Id,V\<rangle>term_rel\<rangle>option_rel\<rangle>nres_rel"
unfolding next_res_wit_refine_def[abs_def] has_res_wit_def
by (autoref_monadic (plain))

concrete_definition next_res_wit_code uses next_res_wit_refine_aux

definition [simp]: "op_next_res_wit rs m \<equiv> SPEC(next_res_wit_inv rs m)"
context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "SPEC(next_res_wit_inv rs m) \<equiv> op_next_res_wit$rs$m"
by simp_all
end

lemma next_res_wit_autoref[autoref_rules]:
  "(\<lambda>rs m. RETURN (next_res_wit_code rs m), op_next_res_wit) \<in>
    \<langle>\<langle>Id,Id\<rangle>ta_rule_rel\<rangle>rs.rel \<rightarrow> \<langle>Id,\<langle>Id,V\<rangle>term_rel\<rangle>rm.rel \<rightarrow> \<langle>\<langle>Id \<times>\<^sub>r \<langle>Id,V\<rangle>term_rel\<rangle>option_rel\<rangle>nres_rel"
proof (intro fun_relI, goal_cases)
  case prems: (1 rs rs' m m')
    hence fin: "finite rs'" by (auto simp: br_def)
    note next_res_wit_code.refine[OF prems, THEN nres_relD]
    also note next_res_wit_refine[OF fin]
    finally show ?case by (auto intro: nres_relI)
qed

lemma [autoref_itype]:
  "op_next_res_wit ::\<^sub>i \<langle>\<langle>F,Q\<rangle>\<^sub>ii_ta_rule\<rangle>\<^sub>ii_set \<rightarrow>\<^sub>i \<langle>Q,\<langle>F,V\<rangle>\<^sub>ii_term\<rangle>\<^sub>ii_map \<rightarrow>\<^sub>i \<langle>\<langle>\<langle>Q,\<langle>F,V\<rangle>\<^sub>ii_term\<rangle>\<^sub>ii_prod\<rangle>\<^sub>ii_option\<rangle>\<^sub>ii_nres"
by simp_all

definition "update_all_refine m ks v \<equiv> FOREACH ks (\<lambda>k m. RETURN (m(k \<mapsto> v))) m"

lemma update_all_ref:
  assumes "finite ks"
  shows "update_all_refine m ks v \<le> RETURN(update_all m ks v)"
unfolding update_all_refine_def
by (refine_vcg FOREACH_rule[OF assms, where I="\<lambda>it r. r = update_all m (ks - it) v"])
   (auto simp: update_all_def fun_eq_iff)

schematic_goal update_all_refine_aux:
  assumes [autoref_rules]: "(M',M) \<in> \<langle>Id,Rv\<rangle>rm.rel"
  assumes [autoref_rules]: "(S',S) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(V',V) \<in> Rv"
  shows "(?f::?'a, update_all_refine M S V) \<in> ?R"
unfolding update_all_refine_def[abs_def]
by (autoref_monadic (plain))

concrete_definition update_all_code uses update_all_refine_aux

lemma update_all_autoref[autoref_rules]:
  assumes "PREFER_id Rk"
  shows "(update_all_code, update_all) \<in> \<langle>Rk,Rv\<rangle>rm.rel \<rightarrow> \<langle>Rk\<rangle>rs.rel \<rightarrow> Rv \<rightarrow> \<langle>Rk,Rv\<rangle>rm.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "Rk = Id" by simp
  case rel: (1 m m' ks ks')
    hence fin: "finite ks'" by (auto simp: br_def)
    note update_all_code.refine[OF rel[unfolded id], THEN nres_relD] 
    also note update_all_ref[OF fin]
    finally show ?case by simp
qed

definition update_all2_refine where
  "update_all2_refine m ks1 ks2 v \<equiv>
    FOREACH ks1
      (\<lambda>k1 m. let m' = update_all (case_option Map.empty id (m k1)) ks2 v in
        RETURN (m(k1 \<mapsto> m'))) m"

lemma update_all2_ref:
  assumes "finite ks1"
  shows "update_all2_refine m ks1 ks2 v \<le> RETURN(update_all2 m ks1 ks2 v)"
unfolding update_all2_refine_def update_all_def
by (refine_vcg FOREACH_rule[OF assms, where I="\<lambda>it r. r = update_all2 m (ks1 - it) ks2 v"])
   (auto simp: update_all2_def fun_eq_iff split: option.split)

schematic_goal update_all2_refine_aux:
  assumes [autoref_rules]: "(M',M) \<in> \<langle>Id,\<langle>Id,Rv\<rangle>rm.rel\<rangle>rm.rel"
  assumes [autoref_rules]: "(S1',S1) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(S2',S2) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(V',V) \<in> Rv"
  shows "(?f::?'a, update_all2_refine M S1 S2 V) \<in> ?R"
unfolding update_all2_refine_def[abs_def]
by (autoref_monadic (plain))

concrete_definition update_all2_code uses update_all2_refine_aux

lemma update_all2_autoref[autoref_rules]:
  assumes "PREFER_id Rk1" and "PREFER_id Rk2"
  shows "(update_all2_code, update_all2) \<in> \<langle>Rk1,\<langle>Rk2,Rv\<rangle>rm.rel\<rangle>rm.rel \<rightarrow> \<langle>Rk1\<rangle>rs.rel \<rightarrow> \<langle>Rk2\<rangle>rs.rel \<rightarrow> Rv \<rightarrow> \<langle>Rk1,\<langle>Rk2,Rv\<rangle>rm.rel\<rangle>rm.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "Rk1 = Id" "Rk2 = Id" by simp_all
  case rel: (1 m m' ks1 ks1' ks2 ks2' v v')
    hence fin: "finite ks1'" by (auto simp: br_def)
    note update_all2_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note update_all2_ref[OF fin]
    finally show ?case by simp
qed

definition map_add2_refine where
  "map_add2_refine m1 m2 \<equiv>
    FOREACH (map_to_set m2)
      (\<lambda>(k,v) m. RETURN (case m k of Some v' \<Rightarrow> m(k \<mapsto> v' ++ v) | None \<Rightarrow> m(k \<mapsto> v))) m1"

lemma map_add2_ref:
  assumes "finite (dom m2)"
  shows "map_add2_refine m1 m2 \<le> RETURN(map_add2 m1 m2)"
proof -
  note rule = FOREACH_rule[where I="\<lambda>it r. r = map_add2 m1 (set_to_map ((map_to_set m2) - it))"]
  show ?thesis unfolding map_add2_refine_def
  proof (refine_vcg rule, simp_all only: it_step_insert_iff, goal_cases)
    case prems: (3 _ it r k v)
      note inj_diff = inj_on_image_set_diff[OF inj_on_fst_map_to_set Diff_subset]
      from prems have fst: "fst (k,v) \<notin> fst ` (map_to_set m2 - it)" by (auto simp: inj_diff)
      note inj = inj_on_fst_map_to_set[THEN inj_on_diff]
      note [simp] = fun_eq_iff map_add2_def Map.map_add_def set_to_map_simp[OF inj] set_to_map_insert[OF fst]
      from fst show ?case by (auto split: option.splits)
  qed (simp_all add: map_to_set_inverse map_add2_def finite_map_to_set assms fun_eq_iff)
qed


schematic_goal map_add2_refine_aux:
  assumes [autoref_rules]: "(M1',M1) \<in> \<langle>Id,\<langle>Id,Rv\<rangle>rm.rel\<rangle>rm.rel"
  assumes [autoref_rules]: "(M2',M2) \<in> \<langle>Id,\<langle>Id,Rv\<rangle>rm.rel\<rangle>rm.rel"
  shows "(?f::?'a, map_add2_refine M1 M2) \<in> ?R"
unfolding map_add2_refine_def[abs_def]
by (autoref_monadic (plain))

concrete_definition map_add2_code uses map_add2_refine_aux

lemma rm_rel_dom_finite:
  assumes id: "Rk = Id"
      and rm: "(c,a) \<in> \<langle>Rk,Rv\<rangle>rm.rel"
    shows "finite (dom a)"
proof -
  from rm id obtain x where "rm.invar c" "x = rm.\<alpha> c"
    and rel: "(x,a) \<in> Id \<rightarrow> \<langle>Rv\<rangle>option_rel" by (auto simp: br_def map.rel_def)
  hence "finite (dom x)" using rm.finite by simp
  also from rel have "dom x = dom a" by (force simp: dom_def elim: option_relE dest: fun_relD)
  finally show ?thesis .
qed

lemma map_add2_autoref[autoref_rules]:
  assumes "PREFER_id Rk1" and "PREFER_id Rk2"
  shows "(map_add2_code, map_add2) \<in> \<langle>Rk1,\<langle>Rk2,Rv\<rangle>rm.rel\<rangle>rm.rel \<rightarrow> \<langle>Rk1,\<langle>Rk2,Rv\<rangle>rm.rel\<rangle>rm.rel \<rightarrow> \<langle>Rk1,\<langle>Rk2,Rv\<rangle>rm.rel\<rangle>rm.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "Rk1 = Id" "Rk2 = Id" by simp_all
  case rel: 1
    note map_add2_code.refine[OF rel[unfolded id], THEN nres_relD]
    also note map_add2_ref[OF rm_rel_dom_finite, OF _ rel(2)]
    finally show ?case by simp
qed

subsection {* Refinement of @{term res_wits} *}

abbreviation "res_wit_map V \<equiv> \<langle>Id,\<langle>Id,V\<rangle>term_rel\<rangle>rm.rel"

schematic_goal res_wits_refine_aux:
  shows "(?f, res_wits) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>res_wit_map V\<rangle>nres_rel"
unfolding res_wits_def[abs_def]
by autoref

concrete_definition res_wits_impl uses res_wits_refine_aux
lemmas [autoref_rules] = res_wits_impl.refine

schematic_goal res_wits_transfer_aux: 
  "RETURN ?c \<le> res_wits_impl TAi"
unfolding res_wits_impl_def 
by (refine_transfer (post))

concrete_definition res_wits_code for TAi uses res_wits_transfer_aux
lemmas [refine_transfer] = res_wits_code.refine
thm res_wits_code_def

schematic_goal ta_only_res_wits_refine_aux:
  shows "(?f, ta_only_res_wits) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> res_wit_map V \<rightarrow> \<langle>Id,Id\<rangle>dflt_TA_rel"
unfolding ta_only_res_wits_def[abs_def]
by autoref

concrete_definition ta_only_res_wits_code uses ta_only_res_wits_refine_aux
lemmas [autoref_rules] = ta_only_res_wits_code.refine

subsection {* Refinement of @{term prs_wits} *}
  
lemma eps_invcls_impl:
  "eps_icls TA qs = op_union_image qs (eps_icl TA)"
by auto

lemma the_Var_autoref:
  "PREFER_id V \<Longrightarrow> PREFER_id F \<Longrightarrow> (the_Var, the_Var) \<in> \<langle>F,V\<rangle>term_rel \<rightarrow> V"
by simp

abbreviation "prs_wit_map \<equiv> \<langle>Id,\<langle>Id,Id\<rangle>ctxt_rel\<rangle>rm.rel"

schematic_goal prs_wits_refine_aux:
  notes [autoref_rules] = the_Var_autoref
  shows "(?f, prs_wits) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>prs_wit_map\<rangle>nres_rel"
unfolding prs_wits_def[abs_def] add_prs_wit_def add_prs_wit_aux_def eps_invcls_impl
by autoref

concrete_definition prs_wits_impl uses prs_wits_refine_aux
lemmas [autoref_rules] = prs_wits_impl.refine

schematic_goal prs_wits_transfer_aux: 
  "RETURN ?c \<le> prs_wits_impl TAi"
unfolding prs_wits_impl_def 
by (refine_transfer (post))

concrete_definition prs_wits_code for TAi uses prs_wits_transfer_aux
lemmas [refine_transfer] = prs_wits_code.refine
thm prs_wits_code_def

schematic_goal ta_only_prs_wits_refine_aux:
  shows "(?f, ta_only_prs_wits) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> prs_wit_map \<rightarrow> \<langle>Id,Id\<rangle>dflt_TA_rel"
unfolding ta_only_prs_wits_def[abs_def]
by autoref

concrete_definition ta_only_prs_wits_code uses ta_only_prs_wits_refine_aux
lemmas [autoref_rules] = ta_only_prs_wits_code.refine

subsection {* Refinement of @{term trim_ta_wits} *}

schematic_goal trim_ta_wits_aux:
  shows "(?f, trim_ta_wits) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>\<langle>Id,Id\<rangle>dflt_TA_rel \<times>\<^sub>r res_wit_map V \<times>\<^sub>r prs_wit_map\<rangle>nres_rel"
unfolding trim_ta_wits_def[abs_def]
by autoref

concrete_definition trim_ta_wits_impl uses trim_ta_wits_aux
lemmas [autoref_rules] = trim_ta_wits_impl.refine

schematic_goal trim_ta_wits_transfer_aux: 
  "RETURN ?c \<le> trim_ta_wits_impl TAi"
unfolding trim_ta_wits_impl_def 
by (refine_transfer (post))

concrete_definition trim_ta_wits_code for TAi uses trim_ta_wits_transfer_aux
lemmas [refine_transfer] = trim_ta_wits_code.refine
thm trim_ta_wits_code_def

subsection {* Refinement of @{term ta_res} *}

definition ta_res_args_aux :: "(_,_)ta \<Rightarrow> _" where
  "ta_res_args_aux TA f qss \<equiv>
    \<Union>((\<lambda>r. case r of TA_rule f' qs q \<Rightarrow>
          if f' = f \<and> list_all2 (\<lambda>qi qsi. qi \<in> qsi) qs qss then eps_cl TA q else {}
      ) ` (ta_rules TA))"

definition ta_res_args_aux_opt :: "(_,_)ta \<Rightarrow> _" where
  "ta_res_args_aux_opt TA f qss \<equiv>
    case ta_idx TA (f, length qss) of
      None \<Rightarrow> {}
    | Some rs \<Rightarrow> \<Union>((\<lambda>(qs,q). if list_all2 (\<lambda>qi qsi. qi \<in> qsi) qs qss then eps_cl TA q else {}) ` rs)"

lemma ta_res_args_aux_opt:
  "ta_res_args_aux TA f qss = ta_res_args_aux_opt TA f qss"
by (auto simp: ta_idx_def ta_res_args_aux_def ta_res_args_aux_opt_def list_all2_conv_all_nth
        split: ta_rule.splits option.splits if_splits) (metis)

schematic_goal ta_res_args_aux_ref_aux:
  assumes "PREFER_id F" and "PREFER_id Q"
  shows "(?f, ta_res_args_aux) \<in> \<langle>F,Q\<rangle>dflt_TA_rel \<rightarrow> F \<rightarrow> \<langle>\<langle>Q\<rangle>rs.rel\<rangle>list_rel \<rightarrow> \<langle>Q\<rangle>rs.rel"
unfolding ta_res_args_aux_opt[abs_def] ta_res_args_aux_opt_def assms[THEN PREFER_id_D]
by autoref

concrete_definition ta_res_args_aux_code uses ta_res_args_aux_ref_aux
lemmas [autoref_rules] = ta_res_args_aux_code.refine

primrec ta_res_impl where
  "ta_res_impl TA (Var q) = eps_cl TA q"
| "ta_res_impl TA (Fun f ts) = (let qss = map (ta_res_impl TA) ts in ta_res_args_aux TA f qss)"

lemma ta_res_impl: "ta_res TA t = ta_res_impl TA t"
by (induction t, auto simp add: ta_res_args_aux_def list_all2_conv_all_nth split: ta_rule.splits if_splits)
    metis

schematic_goal ta_res_impl_aux:
  shows "(?f, ta_res_impl) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>Id,Id\<rangle>term_rel \<rightarrow> \<langle>Id\<rangle>rs.rel"
unfolding ta_res_impl_def[abs_def] 
by autoref

concrete_definition ta_res_code uses ta_res_impl_aux
thm ta_res_code_def[unfolded ta_res_args_aux_code_def]

lemma ta_res_autoref[autoref_rules]:
  assumes "PREFER_id F"
  assumes "PREFER_id Q"
  shows "(ta_res_code, ta_res) \<in> \<langle>F,Q\<rangle>dflt_TA_rel \<rightarrow> \<langle>F,Q\<rangle>term_rel \<rightarrow> \<langle>Q\<rangle>rs.rel"
using assms ta_res_code.refine unfolding ta_res_impl[abs_def] by simp

subsection {* Refinement of @{term ta_match} *}

definition set_App where
"set_App X XS = {z | z x xs. z = x @ xs \<and> x \<in> X \<and> xs \<in> XS}"

definition "set_App_ref X XS \<equiv> FOREACH X (\<lambda>x r. FOREACH XS (\<lambda>xs r. RETURN (insert (x@xs) r)) r) {}"

primrec concat_listset where
"concat_listset [] = {[]}" |
"concat_listset (A # As) = set_App A (concat_listset As)"

lemma concat_listset: "concat ` listset m = concat_listset m"
proof (induction m, auto simp add: set_Cons_def set_App_def, goal_cases)
  case prems: (2 A m x xs)
    let ?P = "\<lambda>x xs. x \<in> A \<and> xs \<in> listset m"
    have "concat ` {x # xs |x xs. ?P x xs} = {concat (x#xs) |x xs. ?P x xs}" by blast
    also have "... = {x@concat xs |x xs. ?P x xs}" by simp
    also have "... = {x@xs | x xs. x \<in> A \<and> xs \<in> concat ` listset m}" by auto
    finally show ?case using prems by auto
qed blast

lemma set_App_ref:
  assumes X: "finite X"
      and XS: "finite XS"
    shows "set_App_ref X XS \<le> RETURN(set_App X XS)"
proof -
  note ruleX = FOREACH_rule[OF X, where I = "\<lambda>itX r. r = set_App (X - itX) XS"]
  show ?thesis unfolding set_App_ref_def
  proof (refine_vcg ruleX, goal_cases)
    case prems: (2 x itX r)
    let ?I = "\<lambda>itXS r. r = set_App (X - itX) XS \<union> set_App (insert x (X - itX)) (XS - itXS)"
    note ruleXS = FOREACH_rule[OF XS, where I = ?I]
    from prems show ?case by (refine_vcg ruleXS, simp_all add: set_App_def, blast+)
  qed (simp_all add: set_App_def)
qed

schematic_goal set_App_ref_aux: 
  assumes [autoref_rules]: "(X',X) \<in> \<langle>\<langle>Id\<rangle>list_rel\<rangle>rs.rel"
  assumes [autoref_rules]: "(XS',XS) \<in> \<langle>\<langle>Id\<rangle>list_rel\<rangle>rs.rel"
  shows "(?f::?'a, set_App_ref X XS) \<in> ?R"
unfolding set_App_ref_def[abs_def] 
by (autoref_monadic (plain))

concrete_definition set_App_code uses set_App_ref_aux

lemma set_App_autoref[autoref_rules]:
  assumes "PREFER_id A"
  shows "(set_App_code, set_App) \<in> \<langle>\<langle>A\<rangle>list_rel\<rangle>rs.rel \<rightarrow> \<langle>\<langle>A\<rangle>list_rel\<rangle>rs.rel \<rightarrow> \<langle>\<langle>A\<rangle>list_rel\<rangle>rs.rel"
proof (intro fun_relI, goal_cases)
  from assms have id[simp]: "A = Id" by simp_all
  case prems: (1 X X' XS XS')
    hence fin: "finite X'" "finite XS'" by (auto simp: br_def)
    note set_App_code.refine[OF prems[unfolded id], THEN nres_relD]
    also note set_App_ref[OF fin]
    finally show ?case by simp
qed

schematic_goal concat_listset_aux:
  assumes [relator_props]:  "A = Id"
  shows "(?f, concat_listset) \<in> \<langle>\<langle>\<langle>Id\<rangle>list_rel\<rangle>rs.rel\<rangle>list_rel \<rightarrow> \<langle>\<langle>Id\<rangle>list_rel\<rangle>rs.rel"
unfolding concat_listset_def[abs_def] 
by autoref

concrete_definition concat_listset_code uses concat_listset_aux

definition [simp]: "op_concat_listset X \<equiv> concat ` listset X"
context begin interpretation autoref_syn .
lemma [autoref_op_pat]: 
  "concat ` listset X \<equiv> op_concat_listset$X"
by simp_all
end

lemma concat_listset_autoref[autoref_rules]:
  assumes "PREFER_id A"
  shows "(concat_listset_code, op_concat_listset) \<in> \<langle>\<langle>\<langle>A\<rangle>list_rel\<rangle>rs.rel\<rangle>list_rel \<rightarrow> \<langle>\<langle>A\<rangle>list_rel\<rangle>rs.rel"
using concat_listset_code.refine[OF PREFER_id_D[OF assms(1)]]
unfolding op_concat_listset_def[abs_def] concat_listset PREFER_id_D[OF assms(1)] .

definition ta_match_var :: "(_,_)ta \<Rightarrow> _" where
  "ta_match_var TA Qsig x Q \<equiv> { [(x,q')] | q'. q' \<in> Qsig \<and> (\<exists> q \<in> Q. (q',q) \<in> (ta_eps TA)\<^sup>*)}"

definition ta_match_var_ref :: "(_,_)ta \<Rightarrow> _" where
  "ta_match_var_ref TA Qsig x Q \<equiv>
    FOREACH Qsig (\<lambda>q res. RETURN (if eps_cl TA q \<inter> Q \<noteq> {} then insert [(x,q)] res else res)) {}"

lemma ta_match_var_ref:
  assumes "finite Qsig"
  shows "ta_match_var_ref TA Qsig x Q \<le> RETURN(ta_match_var TA Qsig x Q)"
unfolding ta_match_var_ref_def
by (refine_vcg FOREACH_rule[OF assms, where I = "\<lambda>it r. r = ta_match_var TA (Qsig - it) x Q"])
   (auto simp: ta_match_var_def)

schematic_goal ta_match_var_ref_aux:
  assumes [autoref_rules]: "(TA',TA) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel"
  assumes [autoref_rules]: "(Qsig',Qsig) \<in> \<langle>Id\<rangle>rs.rel"
  assumes [autoref_rules]: "(x',x) \<in> Id"
  assumes [autoref_rules]: "(Q',Q) \<in> \<langle>Id\<rangle>rs.rel"
  shows "(?f, ta_match_var_ref TA Qsig x Q) \<in> \<langle>\<langle>\<langle>Id \<times>\<^sub>r Id\<rangle>list_rel\<rangle>rs.rel\<rangle>nres_rel"
unfolding ta_match_var_ref_def[abs_def]
by (autoref_monadic (plain))

concrete_definition ta_match_var_ref_code uses ta_match_var_ref_aux

lemma ta_match_var_refine:
  "(ta_match_var_ref_code, ta_match_var) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>Id\<rangle>rs.rel \<rightarrow> Id \<rightarrow> \<langle>Id\<rangle>rs.rel \<rightarrow> \<langle>\<langle>Id \<times>\<^sub>r Id\<rangle>list_rel\<rangle>rs.rel"
proof (intro fun_relI, goal_cases)
  case rel: (1 _ _ Qsig Qsig')
    hence finite: "finite Qsig'" by (simp add: br_def)
    note ta_match_var_ref_code.refine[OF rel, THEN nres_relD]
    also note ta_match_var_ref[OF finite]
    finally show ?case by simp
qed

definition ta_match_fun :: "(_,_)ta \<Rightarrow> _" where
  "ta_match_fun TA f n Q rec \<equiv>
    \<Union>((\<lambda>r. case r of TA_rule f' qs q \<Rightarrow>
          if f = f' \<and> length qs = n \<and> eps_cl TA q \<inter> Q \<noteq> {} then
            concat ` (listset (rec qs))
          else
            {}) ` (ta_rules TA))"

definition ta_match_fun_opt :: "(_,_)ta \<Rightarrow> _" where
  "ta_match_fun_opt TA f n Q rec \<equiv>
    case ta_idx TA (f, n) of
      None \<Rightarrow> {}
    | Some rs \<Rightarrow>
        \<Union>((\<lambda>(qs,q).
            if eps_cl TA q \<inter> Q \<noteq> {} then
              concat ` (listset (rec qs))
            else
              {}) ` rs)"

lemma ta_match_fun_opt:
  "ta_match_fun TA f n Q rec = ta_match_fun_opt TA f n Q rec"
by (auto simp: ta_match_fun_def ta_match_fun_opt_def ta_idx_def
        split: ta_rule.splits option.splits if_splits) blast

schematic_goal ta_match_fun_aux:
  shows "(?f, ta_match_fun) \<in>
    \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> Id \<rightarrow> Id \<rightarrow> \<langle>Id\<rangle>rs.rel \<rightarrow> (\<langle>Id\<rangle>list_rel \<rightarrow> \<langle>\<langle>\<langle>Id \<times>\<^sub>r Id\<rangle>list_rel\<rangle>rs.rel\<rangle>list_rel) \<rightarrow> \<langle>\<langle>Id \<times>\<^sub>r Id\<rangle>list_rel\<rangle>rs.rel"
unfolding ta_match_fun_opt[abs_def] ta_match_fun_opt_def
by autoref

concrete_definition ta_match_fun_code uses ta_match_fun_aux

fun map2 where
  "map2 f [] ys = []"
| "map2 f xs [] = []"
| "map2 f (x#xs) (y#ys) = f x y # map2 f xs ys"

lemma map2_cong[fundef_cong]:
  assumes "xs = xs'" and "\<And>x y. x \<in> set xs' \<Longrightarrow> f x y = g x y"
  shows "map2 f xs = map2 g xs'"
unfolding fun_eq_iff proof (intro allI)
  fix x
  show "map2 f xs x = map2 g xs' x" 
    using assms by (induction f xs x arbitrary: xs' rule: map2.induct) auto
qed

fun ta_match_impl where
  "ta_match_impl TA Qsig (Var x) Q =
    ta_match_var TA Qsig x Q"
| "ta_match_impl TA Qsig (Fun f ts) Q =
    ta_match_fun TA f (length ts) Q (map2 (\<lambda>t Q. ta_match_impl TA Qsig t {Q}) ts)"

fun ta_match_code where
  "ta_match_code TA Qsig (Var x) Q =
    ta_match_var_ref_code TA Qsig x Q"
| "ta_match_code TA Qsig (Fun f ts) Q =
    ta_match_fun_code TA f (length ts) Q (map2 (\<lambda>t Q. ta_match_code TA Qsig t (rs.sng Q)) ts)"

lemma map2_param2_cong:
  assumes "\<And>x y y' . x \<in> set xs \<Longrightarrow> (y,y') \<in> Rb \<Longrightarrow> (f x y, f' x y') \<in> Rc"
      and "(ys,ys') \<in> \<langle>Rb\<rangle>list_rel"
    shows "(map2 f xs ys, map2 f' xs ys') \<in> \<langle>Rc\<rangle>list_rel"           
using assms by (induction f xs ys arbitrary: ys' rule: map2.induct, auto elim!: list_relE)                          

lemma ta_match_code_ref:
  "(ta_match_code, ta_match_impl) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> \<langle>Id\<rangle>rs.rel \<rightarrow> \<langle>Id,Id\<rangle>term_rel \<rightarrow> \<langle>Id\<rangle>rs.rel \<rightarrow> \<langle>\<langle>Id \<times>\<^sub>r Id\<rangle>list_rel\<rangle>rs.rel"
proof (intro fun_relI, goal_cases)
  case (1 _ _ _ _ t t' Q Q')
    thus ?case proof (induction t arbitrary: t' Q Q')
      case (Fun f ts)
        thus ?case
        apply (auto elim!: term_relE simp only: ta_match_code.simps ta_match_impl.simps)
        apply (intro ta_match_fun_code.refine[THEN fun_relD,THEN fun_relD,THEN fun_relD,THEN fun_relD,THEN fun_relD])
        apply (simp_all)
        apply (intro fun_relI map2_param2_cong[where Rb = "Id"])
        apply (parametricity)
        apply (auto simp: br_def rs.correct)
        done
    qed (insert ta_match_var_refine, auto elim!: term_relE dest: fun_relD)
qed

lemma map2_length_eq:
  assumes "length ys = length xs"
    shows "length (map2 g xs ys) = length xs"
using assms by (induction rule: list_induct2, auto)

lemma Suc_lengthE:
  assumes "length xs = Suc (length ys)"
  obtains z zs where "xs = z#zs" "length zs = length ys"
using assms by (cases xs) auto

lemma ta_match_impl:
  fixes TA :: "('q,'f)ta" and t :: "('f,'v)term"
  shows "ta_match_impl TA Qsig t Q = ta_match TA Qsig t Q"
proof (induction t arbitrary: Q)
  case (Fun f ts)
    {
      fix qs :: "'q list" and  xs :: "('v\<times>'q) list list" and i
      assume i: "i < length ts" and lenxs: "length xs = length ts" and lenqs: "length qs = length ts"
      from i Fun.IH have IH: "ta_match TA Qsig (ts!i) {qs!i} = ta_match_impl TA Qsig (ts!i) {qs!i}" by auto
      have "xs!i \<in> map2 (\<lambda>t Q. ta_match_impl TA Qsig t {Q}) ts qs ! i \<longleftrightarrow> xs!i \<in>  ta_match TA Qsig (ts!i) {qs!i}"
      using lenqs lenxs i unfolding IH
      by (induction arbitrary: xs i rule: list_induct2) (auto simp: nth_Cons split: nat.splits elim!: Suc_lengthE)
    } note IH = this
    show ?case by (force simp add: IH ta_match_fun_def listset map2_length_eq split: ta_rule.splits if_splits)
qed (simp add: ta_match_var_def)

lemma ta_match_autoref[autoref_rules]:
  assumes "PREFER_id F" "PREFER_id Q" "PREFER_id V"
  shows "(ta_match_code, ta_match) \<in> \<langle>F,Q\<rangle>dflt_TA_rel \<rightarrow> \<langle>Q\<rangle>rs.rel \<rightarrow> \<langle>F,V\<rangle>term_rel \<rightarrow> \<langle>Q\<rangle>rs.rel \<rightarrow> \<langle>\<langle>V \<times>\<^sub>r Q\<rangle>list_rel\<rangle>rs.rel"
using assms ta_match_code_ref unfolding ta_match_impl[abs_def] by simp

subsection {* Refinement of @{term is_compatible} *}

lemma [autoref_rules, param]:
  "(isOK, isOK) \<in> \<langle>A,B\<rangle>sum_rel \<rightarrow> Id"
unfolding isOK_def[abs_def] by parametricity

lemma fun_of_idref:
  "PREFER_id A \<Longrightarrow> PREFER_id B \<Longrightarrow> (fun_of, fun_of) \<in> \<langle>A \<times>\<^sub>r B\<rangle>list_rel \<rightarrow> A \<rightarrow> B"
by simp

lemma isOK_is_Inr:
  "isOK e = is_Inr e"
by (cases e, simp_all)

abbreviation "term_pair \<equiv> \<langle>Id,Id\<rangle>term_rel \<times>\<^sub>r \<langle>Id,Id\<rangle>term_rel"
abbreviation "wit_tuple \<equiv> Id \<times>\<^sub>r term_pair"
abbreviation "wit_rel \<equiv> \<langle>Id,\<langle>Id,term_pair\<rangle>rm.rel\<rangle>rm.rel"
abbreviation "wit_result \<equiv> \<langle>wit_tuple, wit_rel\<rangle>sum_rel"
abbreviation "trs \<equiv> \<langle>\<langle>Id,Id::('x::compare_order \<times> 'x) set\<rangle>term_rel \<times>\<^sub>r \<langle>Id,Id\<rangle>term_rel\<rangle>rs.rel"

schematic_goal is_compatible_aux:
  notes fun_of_idref[autoref_rules]
  shows "(?f, is_compatible) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> trs \<rightarrow> \<langle>wit_result\<rangle>nres_rel"
unfolding is_compatible_def[abs_def] ta_match'_def isOK_is_Inr
by autoref

concrete_definition is_compatible_impl uses is_compatible_aux
lemmas [autoref_rules] = is_compatible_impl.refine

schematic_goal is_compatible_transfer_aux: 
  "RETURN ?f \<le> is_compatible_impl TA R"
unfolding is_compatible_impl_def
by (refine_transfer (post))

concrete_definition is_compatible_code for TA R uses is_compatible_transfer_aux
lemmas [refine_transfer] = is_compatible_code.refine
thm is_compatible_code_def

subsection {* Refinement of @{term is_coh_final} *}

schematic_goal is_coh_final_aux:
  shows "(?f, is_coh_final) \<in> \<langle>Id\<rangle>rs.rel \<rightarrow> wit_rel \<rightarrow> \<langle>\<langle>term_pair\<rangle>option_rel\<rangle>nres_rel"
unfolding is_coh_final_def[abs_def]
by autoref

concrete_definition is_coh_final_impl uses is_coh_final_aux
lemmas [autoref_rules] = is_coh_final_impl.refine

schematic_goal is_coh_final_transfer_aux: 
  "RETURN ?f \<le> is_coh_final_impl fin rel"
unfolding is_coh_final_impl_def
by (refine_transfer (post))

concrete_definition is_coh_final_code for fin rel uses is_coh_final_transfer_aux
lemmas [refine_transfer] = is_coh_final_code.refine
thm is_coh_final_code_def

subsection {* Refinement of @{term ta_check_comcoh} *}

definition rule_filter_opt where
  "rule_filter_opt TA f qs \<equiv>
    case ta_idx TA (f, length qs) of None \<Rightarrow> {} | Some rs \<Rightarrow> snd ` {r \<in> rs. case r of (qs',q) \<Rightarrow> qs' = qs}"

schematic_goal rule_filter_opt_ref:
  assumes "PREFER_id Q" "PREFER_id F"
  shows "(?f, rule_filter_opt) \<in> \<langle>F,Q\<rangle>dflt_TA_rel \<rightarrow> F \<rightarrow> \<langle>Q\<rangle>list_rel \<rightarrow> \<langle>Q\<rangle>rs.rel"
unfolding rule_filter_opt_def[abs_def] assms[THEN PREFER_id_D]
by autoref

concrete_definition rule_filter_opt_code uses rule_filter_opt_ref
lemmas [autoref_rules] = rule_filter_opt_code.refine

lemma rule_filter_opt:
  "r_rhs ` {r \<in> ta_rules TA. case r of TA_rule f' qs' q' \<Rightarrow> f' = f \<and> qs' = qs} = rule_filter_opt TA f qs"
by (auto simp: rule_filter_opt_def ta_idx_def image_iff r_rhs_unfold split: ta_rule.splits option.splits)+

lemma autoref_the_id:
  "PREFER_id R \<Longrightarrow> (the, the) \<in> \<langle>R\<rangle>option_rel \<rightarrow> R"
using assms by simp

schematic_goal ta_check_comcoh_aux:
  notes param_list_update[autoref_rules]
  notes param_upt[autoref_rules]
  notes autoref_the_id[autoref_rules]
  notes isOK_is_Inr[simp]
  shows "(?f, ta_check_comcoh) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel \<rightarrow> trs \<rightarrow> \<langle>\<langle>term_pair\<rangle>option_rel\<rangle>nres_rel"
unfolding
  ta_check_comcoh_def[abs_def] ta_check_comcoh_trim_def ta_lang_wit_def[abs_def] comp_def[abs_def]
  is_coh_iterate_ref_def[abs_def] new_rel_def isOK_is_Inr rule_filter_opt
by autoref

concrete_definition ta_check_comcoh_impl uses ta_check_comcoh_aux

schematic_goal ta_check_comcoh_transfer_aux: 
  "RETURN ?f \<le> ta_check_comcoh_impl TA R"
unfolding ta_check_comcoh_impl_def
by (refine_transfer (post))

concrete_definition ta_check_comcoh_code for TA R  uses ta_check_comcoh_transfer_aux

no_notation fun_rel_syn (infixr "\<rightarrow>" 60)

subsection \<open>An easier to use interface\<close>

text \<open>
  A dedicated type for tuples that represent tree automata is defined, and useful functions and
  theorems are lifted accordingly. This enables the usage of refined functions without
  relying on Autoref.
\<close>

typedef (overloaded) ('q::compare_order,'f::compare_order) ta_code =
  "{(F,R,E,opt). ((F,R,E,opt), ta.make (rs.\<alpha> F) (rs.\<alpha> R) (ls.\<alpha> E)) \<in> \<langle>Id::(_\<times>'f) set,Id::(_\<times>'q) set\<rangle>dflt_TA_rel}"
proof -
  let ?empty = "(rs.empty(),rs.empty(),ls.empty(),rs.empty(),rm.empty(),True,\<lambda>q. rs.sng q,\<lambda>q. rs.sng q)"
  note simps = TA_rel_def br_def map.rel_def relcomp_unfold  
              rs.correct ls.correct rm.correct
              ta.make_def ta_rhs_states_def ta_det_def ta_idx_def 
  show ?thesis by (intro exI[of _ ?empty]) (auto simp: simps)
qed

setup_lifting type_definition_ta_code

locale ta_code =
 fixes comp_sym :: "'f :: compare_order"
 fixes comp_state :: "'q :: compare_order"
 fixes comp_var :: "'vc :: compare_order"
begin

context begin

private type_synonym 'a rb_set = "('a, unit) rbt"
private type_synonym ('a,'b) rb_ta_rules = "(('a, 'b) ta_rule) rb_set"
private type_synonym ('a,'b) rb_idx = "('b \<times> nat, ('a list \<times> 'a) rb_set) rbt"
private type_synonym 'a rb_eps = "'a \<Rightarrow> 'a rb_set"

private type_synonym ('a,'b) ta_code_tuple =
  "'a rb_set \<times> ('a, 'b) rb_ta_rules \<times> ('a \<times> 'a) dlist \<times>
   'a rb_set \<times> ('a,'b) rb_idx \<times> bool \<times> 'a rb_eps \<times> 'a rb_eps"

private fun ta_abs :: "('q,'f) ta_code_tuple \<Rightarrow> ('q,'f) ta" where
  "ta_abs (F,R,E,_) = ta.make (rs.\<alpha> F) (rs.\<alpha> R) (ls.\<alpha> E)"

private lemma make_in_relI:
  assumes "rs.\<alpha> F' = rs.\<alpha> F" "rs.\<alpha> R' = rs.\<alpha> R" "ls.\<alpha> E' = ls.\<alpha> E"
      and mc: "ta_make_code F R E = x"
    shows "(x, ta.make (rs.\<alpha> F') (rs.\<alpha> R') (ls.\<alpha> E')) \<in> \<langle>Id, Id\<rangle>dflt_TA_rel"
using assms
by (auto simp: mc[symmetric] br_def intro!: ta_make_autoref[THEN fun_relD, THEN fun_relD, THEN fun_relD])

lift_definition \<alpha> :: "('q,'f)ta_code \<Rightarrow> ('q,'f) ta" is ta_abs .
lift_definition res :: "('q,'f)ta_code \<Rightarrow> ('f, 'q) term \<Rightarrow> 'q rb_set" is ta_res_code .
lift_definition make :: "'q rb_set  \<Rightarrow> ('q, 'f) rb_ta_rules  \<Rightarrow> ('q \<times> 'q) dlist \<Rightarrow> ('q,'f)ta_code" is ta_make_code
  by (auto intro!: make_in_relI simp: ta_make_code_def Let_def split: prod.splits)
lift_definition res_wits :: "('q,'f)ta_code \<Rightarrow> ('q, ('f, 'v) term) rbt" is res_wits_code .
lift_definition only_res_wits :: "('q,'f)ta_code \<Rightarrow> ('q, ('f, 'v) term) rbt \<Rightarrow> ('q,'f)ta_code" is ta_only_res_wits_code
  by (auto intro!: make_in_relI simp: ta_only_res_wits_code_def ta_make_code_def Let_def split: prod.splits)
lift_definition prs_wits :: "('q,'f)ta_code \<Rightarrow> ('q, ('f, 'q) ctxt) rbt" is prs_wits_code .
lift_definition only_prs_wits :: "('q,'f)ta_code \<Rightarrow> ('q, ('f, 'q) ctxt) rbt \<Rightarrow> ('q,'f)ta_code" is ta_only_prs_wits_code
  by (auto intro!: make_in_relI simp: ta_only_prs_wits_code_def ta_make_code_def Let_def split: prod.splits)
lift_definition det :: "('q,'f)ta_code \<Rightarrow> bool" is ta_det_impl .
lift_definition check_comcoh_wit :: "('q,'f)ta_code  \<Rightarrow> (('f,'vc)term \<times> ('f,'vc)term) rb_set \<Rightarrow> (('f,'vc)term \<times> ('f,'vc) term) option" is ta_check_comcoh_code .
lift_definition final :: "('q,'f)ta_code \<Rightarrow> 'q rb_set" is ta_final_impl .

definition "make_ls F R E \<equiv> make (rs.from_list F) (rs.from_list R) (ls.from_list E)"

definition "check_comcoh TA R = (check_comcoh_wit TA R = None)"
definition "check_comcoh_wit_ls TA R \<equiv> check_comcoh_wit TA (rs.from_list R)"
definition "check_comcoh_ls TA R = (check_comcoh_wit_ls TA R = None)"

definition "trim_wits TA \<equiv> let res = res_wits TA; TA = only_res_wits TA res; 
                                prs = prs_wits TA; TA = only_prs_wits TA prs in (TA, res, prs)"
definition "trim TA \<equiv> fst (trim_wits TA :: _ \<times> (_,(_,unit)term)rbt \<times> _)"

definition "only_reach TA = only_res_wits TA (res_wits TA :: (_,(_,unit)term)rbt)"
definition "only_prod TA = only_prs_wits TA (prs_wits TA)"

definition empty_wit where "empty_wit TA \<equiv> rm.sel (res_wits TA) (\<lambda>(q,t::('f,'v)term). rs.memb q (final TA))"
definition empty where "empty TA \<equiv> empty_wit TA = (None::(_ \<times> (_,unit)term)option)"

private lemma make:
  "\<alpha> (make F R E) = ta.make (rs.\<alpha> F) (rs.\<alpha> R) (ls.\<alpha> E)"
unfolding \<alpha>.rep_eq make.rep_eq ta_make_code_def
by (auto simp: Let_def split: prod.splits)

private lemma make_ls:
  "\<alpha> (make_ls F R E) = ta.make (set F) (set R) (set E)"
unfolding make_ls_def using make[of "rs.from_list F" "rs.from_list R"]
by (auto simp: rs.correct ls.correct)

private lemma rep_abs_TA_rel:
  "(Rep_ta_code TA, \<alpha> TA) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel"
proof -
  from Rep_ta_code[of TA] obtain F R E opt where
    rep: "Rep_ta_code TA = (F,R,E,opt)" and
    rel: "((F, R, E, opt), ta.make (rs.\<alpha> F) (rs.\<alpha> R) (ls.\<alpha> E)) \<in> \<langle>Id, Id\<rangle>dflt_TA_rel" by auto
  thus ?thesis by (simp add: \<alpha>.rep_eq)
qed

private lemma det:
  "det TA = ta_det (\<alpha> TA)"
proof -
  have "(ta_det_impl (Rep_ta_code TA), ta_det (\<alpha> TA)) \<in> bool_rel"
    by (intro ta_impl_autoref(6)[THEN fun_relD]) (auto intro: rep_abs_TA_rel)
  thus ?thesis unfolding det.rep_eq by simp
qed

private lemma final:
  "rs.\<alpha> (final TA) = ta_final (\<alpha> TA)"
proof -
  have "(ta_final_impl (Rep_ta_code TA), ta_final (\<alpha> TA)) \<in> \<langle>Id\<rangle>rs.rel"
    by (intro ta_impl_autoref(1)[THEN fun_relD]) (auto intro: rep_abs_TA_rel)
  thus ?thesis unfolding final.rep_eq by (simp add: br_def)
qed

private lemma res:
  "rs.\<alpha> (res TA t) = ta_res (\<alpha> TA) t"
proof -
  have "(ta_res_code (Rep_ta_code TA) t, ta_res (\<alpha> TA) t) \<in> \<langle>Id\<rangle>rs.rel"
    by (intro ta_res_autoref[of Id Id, THEN fun_relD, THEN fun_relD]) (auto intro: rep_abs_TA_rel)
  thus ?thesis unfolding res.rep_eq by (auto simp: br_def)
qed

private lemma res_wits_code:
  assumes "(I,A) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel" shows
    "dom (rm.\<alpha> (res_wits_code I)) = ta_reachable A" (is "?A")
    "rm.\<alpha> (res_wits_code I) q = Some t \<Longrightarrow> is_ad_res_wit A q t" (is "?B \<Longrightarrow> ?C")
proof -
  from assms have finite: "ta_finite A" by (auto simp: TA_rel_def br_def ta_finite_def)
  note res_wits_code.refine[of I]
  also note res_wits_impl.refine[THEN fun_relD, THEN nres_relD, OF assms, of Id]
  also note res_wits_correct[OF finite]
  finally have "res_wits_inv A ({}, rm.\<alpha> (res_wits_code I))" by (auto elim!: RETURN_ref_SPECD simp: br_def)
  thus ?A "?B \<Longrightarrow> ?C" by (auto simp: res_wits_inv_def map_to_set_def)
qed

private lemma res_wits:
  "dom (rm.\<alpha> (res_wits TA :: ('q, ('f, 'v) term) rbt)) = ta_reachable (\<alpha> TA)"
  "rm.\<alpha> (res_wits TA) q = Some t \<Longrightarrow> is_ad_res_wit (\<alpha> TA) q (t :: ('f, 'v) term)"
unfolding res_wits.rep_eq by (intro res_wits_code[of "Rep_ta_code TA" "\<alpha> TA"] rep_abs_TA_rel)+

private lemma ta_absI:
  "(I,A) \<in> \<langle>Id, Id\<rangle>dflt_TA_rel \<Longrightarrow> ta_abs I = A"
by (cases I, auto simp: TA_rel_def ta.make_def br_def)

private lemma only_res_wits:
  assumes "dom (rm.\<alpha> wits) = ta_reachable (\<alpha> TA)"
    shows "\<alpha> (only_res_wits TA wits) = ta_only_reach (\<alpha> TA)"
proof -
  note ref = ta_only_res_wits_code.refine[THEN fun_relD, THEN fun_relD, OF rep_abs_TA_rel,
               of wits "rm.\<alpha> wits" Id TA, unfolded ta_only_res_wits[OF assms]]
  show ?thesis unfolding \<alpha>.rep_eq[of "only_res_wits TA wits"] only_res_wits.rep_eq
    by (intro ta_absI ref, simp add: br_def)
qed
  
private lemma prs_wits_code:
  assumes "(I,A) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel" shows
    "dom (rm.\<alpha> (prs_wits_code I)) = ta_productive A" (is "?A")
    "rm.\<alpha> (prs_wits_code I) q = Some C \<Longrightarrow> is_prs_wit A q C" (is "?B \<Longrightarrow> ?C")
proof -
  from assms have finite: "ta_finite A" by (auto simp: TA_rel_def br_def ta_finite_def)
  note prs_wits_code.refine[of I]
  also note prs_wits_impl.refine[THEN fun_relD, THEN nres_relD, OF assms]
  also note prs_wits_correct[OF finite]
  finally have "prs_wits_spec A (rm.\<alpha> (prs_wits_code I))" by (auto elim!: RETURN_ref_SPECD simp: br_def)
  thus ?A "?B \<Longrightarrow> ?C" by (auto simp: res_wits_inv_def map_to_set_def)
qed

private lemma prs_wits:
  "dom (rm.\<alpha> (prs_wits TA)) = ta_productive (\<alpha> TA)" (is "?A")
  "rm.\<alpha> (prs_wits TA) q = Some C \<Longrightarrow> is_prs_wit (\<alpha> TA) q C" (is "?B \<Longrightarrow> ?C")
unfolding prs_wits.rep_eq by (intro prs_wits_code[of "Rep_ta_code TA" "\<alpha> TA"] rep_abs_TA_rel)+

private lemma only_prs_wits:
  assumes "dom (rm.\<alpha> wits) = ta_productive (\<alpha> TA)"
    shows "\<alpha> (only_prs_wits TA wits) = ta_only_prod (\<alpha> TA)"
proof -
  note ref =  ta_only_prs_wits_code.refine[THEN fun_relD, THEN fun_relD, OF rep_abs_TA_rel,
                of wits "rm.\<alpha> wits" TA, unfolded ta_only_prs_wits[OF assms]]
  show ?thesis unfolding \<alpha>.rep_eq[of "only_prs_wits TA wits"] only_prs_wits.rep_eq
    by (intro ta_absI ref, simp add: br_def)
qed

private lemma ta_check_comcoh_code:
  assumes TA: "(I,A) \<in> \<langle>Id,Id\<rangle>dflt_TA_rel"
     and det: "ta_det A"
     and wf: "\<forall>(l, r) \<in> rs.\<alpha> R. vars_term r \<subseteq> vars_term l" shows
  "case (ta_check_comcoh_code I R) of
     None \<Rightarrow> rstep (rs.\<alpha> R) `` ta_lang A \<subseteq> ta_lang A
   | Some (wl, wr) \<Rightarrow> (wl, wr) \<in> rstep (rs.\<alpha> R) \<and> wl \<in> ta_lang A \<and> wr \<notin> ta_lang A"
proof -
  from assms have finiteA: "ta_finite A" by (auto simp: TA_rel_def br_def ta_finite_def)
  have finiteR: "finite (rs.\<alpha> R)" by simp
  note ta_check_comcoh_code.refine[of I R]
  also note ta_check_comcoh_impl.refine[THEN fun_relD, THEN fun_relD, THEN nres_relD, OF TA, of R "rs.\<alpha> R"]
  also note ta_check_comcoh_correct[OF finiteA finiteR det]
  finally show ?thesis using wf by (auto elim!: RETURN_ref_SPECD simp: br_def)
qed

private lemma check_comcoh_wit:
  assumes det: "ta_det (\<alpha> TA)"
      and wf: "\<forall>(l, r) \<in> rs.\<alpha> R. vars_term r \<subseteq> vars_term l" shows
  "case (check_comcoh_wit TA R) of
     None \<Rightarrow> rstep (rs.\<alpha> R) `` ta_lang (\<alpha> TA) \<subseteq> ta_lang (\<alpha> TA)
   | Some (wl, wr) \<Rightarrow> (wl, wr) \<in> rstep (rs.\<alpha> R) \<and> wl \<in> ta_lang (\<alpha> TA) \<and> wr \<notin> ta_lang (\<alpha> TA)"
unfolding check_comcoh_wit.rep_eq
by (intro ta_check_comcoh_code[of "Rep_ta_code TA" "\<alpha> TA" R] assms rep_abs_TA_rel)+

private lemma check_comcoh:
  assumes det: "ta_det (\<alpha> TA)"
     and wf: "\<forall>(l, r) \<in> rs.\<alpha> R. vars_term r \<subseteq> vars_term l" shows
  "check_comcoh TA R = (rstep (rs.\<alpha> R) `` ta_lang (\<alpha> TA) \<subseteq> ta_lang (\<alpha> TA))"
unfolding check_comcoh_def using check_comcoh_wit[of TA R, OF assms]
by (cases "check_comcoh_wit TA R") auto

private lemma check_comcoh_wit_ls:
  assumes det: "ta_det (\<alpha> TA)"
      and wf: "\<forall>(l, r) \<in> set R. vars_term r \<subseteq> vars_term l" shows
  "case (check_comcoh_wit_ls TA R) of
     None \<Rightarrow> rstep (set R) `` ta_lang (\<alpha> TA) \<subseteq> ta_lang (\<alpha> TA)
   | Some (wl, wr) \<Rightarrow> (wl, wr) \<in> rstep (set R) \<and> wl \<in> ta_lang (\<alpha> TA) \<and> wr \<notin> ta_lang (\<alpha> TA)"
unfolding check_comcoh_wit_ls_def using check_comcoh_wit[of TA "rs.from_list R"] assms 
by (auto simp: rs.correct, unfold rs.correct(35)) blast

private lemma check_comcoh_ls:
  assumes det: "ta_det (\<alpha> TA)"
     and wf: "\<forall>(l, r) \<in> set R. vars_term r \<subseteq> vars_term l" shows
  "check_comcoh_ls TA R = (rstep (set R) `` ta_lang (\<alpha> TA) \<subseteq> ta_lang (\<alpha> TA))"
unfolding check_comcoh_ls_def using check_comcoh_wit_ls[of TA R, OF assms]
by (cases "check_comcoh_wit_ls TA R") auto

private lemma trim_wits:
  "case trim_wits TA of (TA', rs :: ('q, ('f, 'v) term) rbt, ps)
    \<Rightarrow> \<alpha> TA' = trim_ta (\<alpha> TA) \<and>
      (\<forall>q \<in> ta_states (\<alpha> TA'). is_ad_res_wit (\<alpha> TA') q (the (rm.\<alpha> rs q))) \<and>
      (\<forall>q \<in> ta_states (\<alpha> TA'). is_prs_wit (\<alpha> TA') q (the (rm.\<alpha> ps q)))"
proof -
  note rw = res_wits[where ?'v = 'v]
  let ?rw = "res_wits TA :: ('q, ('f, 'v) term) rbt"
  let ?pw = "prs_wits (only_res_wits TA ?rw)"
  let ?tar = "ta_only_reach (\<alpha> TA)"
  let ?tapr = "ta_only_prod ?tar"
  let ?states = "ta_states (?tapr)"
  have prod: "\<And>TA. ta_states (ta_only_prod TA) \<subseteq> ta_productive TA" by (auto simp: ta_states_def r_states_def ta_restrict_def)

  from rw[of TA] trim_ta_reachable[of "\<alpha> TA"] have "?states \<subseteq> dom (rm.\<alpha> ?rw)" by auto
  hence "\<forall>q \<in> ?states. is_ad_res_wit ?tar q (the (rm.\<alpha> ?rw q))" by (auto simp: ta_only_reach_res_wit rw(2))
  with prod[of ?tar] ta_res_only_prod[of _ ?tar] have
    rs: "\<forall>q \<in> ?states. is_ad_res_wit ?tapr q (the (rm.\<alpha> ?rw q))" by auto

  from prs_wits(2) have "\<forall>q \<in> dom (rm.\<alpha> ?pw). is_prs_wit (\<alpha> (only_res_wits TA ?rw)) q (the (rm.\<alpha> ?pw q))" by auto
  hence "\<forall>q \<in> dom (rm.\<alpha> ?pw). is_prs_wit ?tar q (the (rm.\<alpha> ?pw q))" by (simp add: only_res_wits[OF rw(1)])
  hence "\<forall>q \<in> ta_productive ?tar. is_prs_wit ?tar q (the (rm.\<alpha> ?pw q))"
    unfolding prs_wits(1)[of "only_res_wits TA ?rw", unfolded only_res_wits[OF rw(1)]] .
  with prod[of ?tar] have ps: "\<forall>q \<in> ?states. is_prs_wit ?tapr q (the (rm.\<alpha> ?pw q))" by (auto simp: ta_only_prod_prs_wit)
  
  note simps = trim_wits_def Let_def res_wits only_res_wits prs_wits only_prs_wits trim_ta_def rs ps
  show ?thesis by (simp add: simps)
qed

private lemma trim:
  "\<alpha> (trim TA) = trim_ta (\<alpha> TA)"
unfolding trim_def using trim_wits[of TA, where ?'v = unit]
by (cases "trim_wits TA :: _ \<times> (_,(_,unit)term)rbt \<times> _") (simp)

private lemma only_reach:
  "\<alpha> (only_reach TA) = ta_only_reach (\<alpha> TA)"
unfolding only_reach_def using only_res_wits[OF res_wits(1)[where ?'v = unit]] by simp

private lemma only_prod:
  "\<alpha> (only_prod TA) = ta_only_prod (\<alpha> TA)"
unfolding only_prod_def using only_prs_wits[OF prs_wits(1)] by simp

private lemma empty_wit:
  "case empty_wit TA of
    None \<Rightarrow> ta_empty (\<alpha> TA)
  | Some (q,t) \<Rightarrow> q \<in> ta_final (\<alpha> TA) \<and> is_ad_res_wit (\<alpha> TA) q (t::('f,'v)term)"
proof -
  let ?empty = "empty_wit TA :: ('q \<times> ('f, 'v) term) option"
  show ?thesis proof (cases "?empty")
    case None
      {
        fix q and t :: "('f,'v)term"
        from None have "rm.\<alpha> (res_wits TA) q = Some t \<Longrightarrow> \<not> (q \<in> ta_final (\<alpha> TA))"
          by (auto simp: empty_wit_def rs.correct final dest!: rm.sel'_noneD[simplified, where u = q and v = t])
      }
      with res_wits(1) have "ta_final (\<alpha> TA) \<inter> ta_reachable (\<alpha> TA) = {}" by (auto simp: dom_def)
      with None show ?thesis by (auto simp: ta_empty_def) next
    case (Some qt)
      obtain q and t :: "('f,'v)term" where [simp]: "qt = (q,t)" by (cases qt)
      from Some have "rm.\<alpha> (res_wits TA) q = Some t" "q \<in> ta_final (\<alpha> TA)"
        by (auto simp: empty_wit_def rs.correct final dest!: rm.sel'_SomeD[simplified])
      with Some res_wits show ?thesis by simp
  qed
qed

private lemma empty:
  "empty TA = ta_empty (\<alpha> TA)"
unfolding empty_def using empty_wit[of TA, where ?'v = unit]
by (cases "empty_wit TA ::(_ \<times> (_,unit)term)option")
   (auto simp: ta_empty[where ?'c = unit] ta_lang_def2)

lemmas correct = make make_ls det final res empty only_reach only_prod trim check_comcoh check_comcoh_ls
lemmas correct_empty_wit = empty_wit
lemmas correct_res_wits = res_wits only_res_wits
lemmas correct_prs_wits = prs_wits only_prs_wits
lemmas correct_trim_wits = trim_wits
lemmas correct_comcoh_wit = check_comcoh_wit check_comcoh_wit_ls

end
end

lemmas [code] =
    ta_code.res_wits.rep_eq
    ta_code.prs_wits.rep_eq
    ta_code.res.rep_eq
    ta_code.det.rep_eq
    ta_code.final.rep_eq
    ta_code.check_comcoh_wit.rep_eq  
    
    ta_code.make_ls_def
    ta_code.check_comcoh_def
    ta_code.check_comcoh_ls_def
    ta_code.check_comcoh_wit_ls_def
    ta_code.trim_def
    ta_code.trim_wits_def
    ta_code.empty_def
    ta_code.empty_wit_def
    ta_code.only_reach_def
    ta_code.only_prod_def

lemmas [code abstract] =
  ta_code.make.rep_eq
  ta_code.only_res_wits.rep_eq
  ta_code.only_prs_wits.rep_eq

end